#
# Authors: Robert Abram <robert.abram@entpack.com>
#
# Copyright (C) 2015-2016 EntPack
# see file 'LICENSE' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import ast
import hashlib
import logging
import socket
import sys
import time

from silentdune_client.utils.module_loading import import_by_str
from silentdune_client.utils.misc import is_service_running
from silentdune_client.modules import BaseModule, QueueTask
from silentdune_client.modules.firewall.manager.slots import Slots
from silentdune_client.modules.firewall.manager import SilentDuneClientFirewallModule, \
    TASK_FIREWALL_INSERT_RULES, TASK_FIREWALL_DELETE_SLOT, TASK_FIREWALL_ALLOW_ALL_DNS_ACCESS, \
    TASK_FIREWALL_DISABLE_ALL_DNS_ACCESS, TASK_FIREWALL_DISABLE_ALL_NTP_ACCESS, TASK_FIREWALL_ALLOW_ALL_NTP_ACCESS
from silentdune_client.builders import iptables as ipt
from silentdune_client.modules.firewall.manager.iptables_utils import resolve_hostname

_logger = logging.getLogger('sd-client')

module_list = {
    'Silent Dune Auto Service Discovery': {
        'module': 'SilentDuneAutoServiceDiscoveryModule',
    },
}

AUTO_DISCOVERY_SERVICE_LIST = (
    "silentdune_client.modules.automation.auto_discovery.dns.DynamicNameServiceDiscovery",  # Slot 130
    "silentdune_client.modules.automation.auto_discovery.ntp.NetworkTimeProtocolDiscovery",  # Slot 140
    "silentdune_client.modules.automation.auto_discovery.dhcp.DynamicHostConfigurationProtocolDiscovery",  # Slot 170
    "silentdune_client.modules.automation.auto_discovery.icmp.IcmpDiscovery",  # Slot 180
    "silentdune_client.modules.automation.auto_discovery.updates.SystemUpdatesDiscovery",  # Slot 150
    # Identity services (LDAP, FreeIPA, ...)  # Slot 160
)

NTP_DISCOVERY_SERVICE = "silentdune_client.modules.automation.auto_discovery.ntp.NetworkTimeProtocolDiscovery"


class SilentDuneAutoServiceDiscoveryModule(BaseModule):
    """
    Auto discover required external services by this system. IE: DNS, NTP, Updates, DHCP, ...
    """
    # Module properties
    _disable_auto_dns = False
    _disable_auto_ntp = False
    _disable_auto_updates = False
    _disable_auto_updates_ftp = False
    _disable_auto_dhcp = False
    _disable_auto_icmp = False

    # Timed events
    _t_all_service_check = 0  # Counter for next all service check
    _t_all_check_interval = 3600  # One hour

    _t_ntp_service_check = 0  # Counter for next NTP service check
    _t_ntp_check_interval = 300  # Five minutes

    _startup = True
    _all_dns_access_enabled = False
    _all_ntp_access_enabled = False

    _flat_rules = None  # Used in the flatten_rules() method

    priority = 2  # High priority startup.

    def __init__(self):

        self._arg_name = 'autodiscover'  # Set argparser name
        self._config_section_name = 'auto_discovery'  # Set configuration file section name

        # Enable multi-threading
        self.wants_processing_thread = True

        self._enabled = False

    def add_installer_arguments(self, parser):
        """
        Virtual Override
        Add our module's argparser arguments
        """

        # Create a argument group for our module
        group = parser.add_argument_group('auto discovery module', 'Silent Dune Auto Discovery module')

        group.add_argument('--discovery-mod-enable', action='store_true',  # noqa
                           help=_('Enable the auto discovery module'))  # noqa

        group.add_argument(_('--disable-auto-dns'), help=_('Disable auto DNS discovery.'),  # noqa
                           default=False, action='store_true')  # noqa
        group.add_argument(_('--disable-auto-ntp'), help=_('Disable auto NTP discovery.'),  # noqa
                           default=False, action='store_true')  # noqa
        group.add_argument(_('--disable-auto-updates'), help=_('Disable auto System Updates discovery.'),  # noqa
                           default=False, action='store_true')  # noqa
        group.add_argument(_('--disable-auto-updates-ftp'), help=_('Disable FTP rule creation for System Updates.'),  # noqa
                           default=False, action='store_true')  # noqa
        group.add_argument(_('--disable-auto-dhcp'), help=_('Disable auto DHCP discovery.'),  # noqa
                           default=False, action='store_true')  # noqa
        group.add_argument(_('--disable-auto-icmp'), help=_('Disable auto ICMP discovery.'),  # noqa
                           default=False, action='store_true')  # noqa

    def validate_arguments(self, args):
        """
        Virtual Override
        Validate command line arguments and save values to our configuration object.
        :param args: An argparse object.
        """
        # See if we have been enabled or not
        if '--discovery-mod-enable' not in sys.argv:
            return True

        if '--disable-auto-updates' in sys.argv and (
                '--disable-auto-updates-ftp' in sys.argv):
            print('sdc-install: argument --disable-auto-updates conflicts with other auto discover module arguments.')
            return False

        if args.discovery_mod_enable:
            self._enabled = args.discovery_mod_enable

        # Check for any required arguments here

        self._disable_auto_dns = args.disable_auto_dns
        self._disable_auto_ntp = args.disable_auto_ntp
        self._disable_auto_updates = args.disable_auto_updates
        self._disable_auto_updates_ftp = args.disable_auto_updates_ftp
        self._disable_auto_dhcp = args.disable_auto_dhcp
        self._disable_auto_icmp = args.disable_auto_icmp

        return True

    def validate_config(self, config):
        """
        Virtual Override
        Validate configuration file arguments and save values to our config object.
        :param config: A ConfigParser object.
        """

        # See if we are enabled or not
        try:
            self._enabled = True if config.get(self._config_section_name, 'enabled').lower() == 'yes' else False
        except:
            _logger.debug('{0} configuration section not found in configuration file.'.format(
                self._config_section_name))
            self._enabled = False

        # Only worry about the rest of the configuration items if we are enabled.
        if self._enabled:

            self._disable_auto_dns = \
                True if config.get(self._config_section_name, 'disable_auto_dns').lower() == 'yes' else False
            self._disable_auto_ntp = \
                True if config.get(self._config_section_name, 'disable_auto_ntp').lower() == 'yes' else False
            self._disable_auto_updates = \
                True if config.get(self._config_section_name, 'disable_auto_updates').lower() == 'yes' else False
            self._disable_auto_updates_ftp = \
                True if config.get(self._config_section_name, 'disable_auto_updates_ftp').lower() == 'yes' else False
            self._disable_auto_dhcp = \
                True if config.get(self._config_section_name, 'disable_auto_dhcp').lower() == 'yes' else False
            self._disable_auto_icmp = \
                True if config.get(self._config_section_name, 'disable_auto_icmp').lower() == 'yes' else False

        return True

    def prepare_config(self, config):
        """
        Virtual Override
        Return the configuration file structure. Any new configuration items should be added here.
        Note: The order should be reverse of the expected order in the configuration file.
        """

        config.set(self._config_section_name, 'enabled', 'yes' if self._enabled else 'no')
        config.set(self._config_section_name, 'disable_auto_dns', 'yes' if self._disable_auto_dns else 'no')
        config.set(self._config_section_name, 'disable_auto_ntp', 'yes' if self._disable_auto_ntp else 'no')
        config.set(self._config_section_name, 'disable_auto_updates', 'yes' if self._disable_auto_updates else 'no')
        config.set(self._config_section_name, 'disable_auto_updates_ftp',
                   'yes' if self._disable_auto_updates_ftp else 'no')
        config.set(self._config_section_name, 'disable_auto_dhcp', 'yes' if self._disable_auto_dhcp else 'no')
        config.set(self._config_section_name, 'disable_auto_icmp', 'yes' if self._disable_auto_icmp else 'no')

        config.set_comment(self._config_section_name, self._config_section_name,
                           _('; Silent Dune Auto Discovery Module Configuration\n'))  # noqa
        config.set_comment(self._config_section_name, 'disable_auto_dns',
                           _('; Disable the auto DNS rule generation.\n'))  # noqa
        config.set_comment(self._config_section_name, 'disable_auto_ntp',
                           _('; Disable the auto NTP rule generation.\n'))  # noqa
        config.set_comment(self._config_section_name, 'disable_auto_updates',
                           _('; Disable the auto System Update rule generation.\n'))  # noqa
        config.set_comment(self._config_section_name, 'disable_auto_updates_ftp',
                           _('; Disable the auto System Update FTP rule generation.\n'))  # noqa
        config.set_comment(self._config_section_name, 'disable_auto_dhcp',
                           _('; Disable the auto DHCP rule generation.\n'))  # noqa
        config.set_comment(self._config_section_name, 'disable_auto_icmp',
                           _('; Disable the auto ICMP rule generation.\n'))  # noqa

        return True

    def service_startup(self):
        _logger.debug('{0}: module startup called'.format(self.get_name()))

        if not self.validate_config(self.config):
            _logger.error('{0}: module configuration validation failed.'.format(self.get_name()))
            return False

        if not self._enabled:
            return True

        return True

    def service_shutdown(self):
        _logger.debug('{0}: module shutdown called'.format(self.get_name()))

    def process_loop(self):

        if self._startup:

            # Try resolving example.org, if we get an error try enabling generic all DNS access.
            # This block may be called multiple times during startup to make sure DNS is working.
            try:
                resolve_hostname('example.org', ipt.TRANSPORT_AUTO)
            except socket.gaierror:

                # If we have reached here, we have no DNS access at all.
                if self._all_dns_access_enabled:
                    raise OSError('No external DNS access available. Unable to complete auto discovery.')

                _logger.debug('{0}: Asking Firewall Module to enable generic DNS access.'.format(self.get_name()))

                # Tell the firewall manager to enable generic all DNS access.
                self._all_dns_access_enabled = True

                task = QueueTask(TASK_FIREWALL_ALLOW_ALL_DNS_ACCESS,
                                 src_module=self.get_name(),
                                 dest_module=SilentDuneClientFirewallModule().get_name())
                self.send_parent_task(task)

                return

            self._startup = False
            self.discover_services()

            # Tell the firewall manager to remove generic all DNS access rules.
            if self._all_dns_access_enabled:
                self._all_dns_access_enabled = False
                _logger.debug('{0}: Asking Firewall Module to disable generic DNS access.'.format(self.get_name()))
                task = QueueTask(TASK_FIREWALL_DISABLE_ALL_DNS_ACCESS,
                                 src_module=self.get_name(),
                                 dest_module=SilentDuneClientFirewallModule().get_name())
                self.send_parent_task(task)

        # After the check interval has passed, check services again.
        if self.timed_event('_t_all_service_check', self._t_all_check_interval):
            self.discover_services()

        # See we need to remove the all access ntp rules.
        if self._all_ntp_access_enabled and self.timed_event('_t_ntp_service_check', self._t_ntp_check_interval):

            rule_count = self.check_service(NTP_DISCOVERY_SERVICE)

            # If we found NTP rules, tell the firewall manager to remove generic all NTP access rules.
            if self._all_ntp_access_enabled and rule_count > 0:

                self._all_ntp_access_enabled = False
                _logger.debug('{0}: Asking Firewall Module to disable generic NTP access.'.format(self.get_name()))
                task = QueueTask(TASK_FIREWALL_DISABLE_ALL_NTP_ACCESS,
                                 src_module=self.get_name(),
                                 dest_module=SilentDuneClientFirewallModule().get_name())
                self.send_parent_task(task)

    def discover_services(self):
        """
        Loop through the auto discovery class list and call discover() for each class found.
        If any firewall rules are found, send them to the firewall manager service.

        Note: Each slot id used in each discovery class needs to be unique.
        """
        for name in AUTO_DISCOVERY_SERVICE_LIST:
            self.check_service(name)

    def check_service(self, name):
        """
        Check the service for rules and add them to the firewall.
        :param name: Service discovery module name
        """
        module_name, class_name = name.rsplit('.', 1)

        _logger.debug('{0}: Loading auto discover object {1}'.format(self.get_name(), class_name))

        module = import_by_str(name)
        cls = module(config=self.config)
        disabled = getattr(self, cls.get_config_property_name())
        if type(disabled) is str:  # Python 2.7 returns string type from getattr(), Python 3.4 returns bool.
            disabled = ast.literal_eval(disabled)

        # _logger.debug('Property: {0}: Value: {1}'.format(cls.get_config_property_name(), disabled))
        # See if this discovery service has been disabled. Name value must match one of our property names.
        if disabled:
            _logger.debug('{0}: {1} discovery service disabled by config.'.format(self.get_name(), class_name))
            return 0

        rules, slot = cls.discover(self)

        rules = self.flatten_rules(rules)

        if rules:

            # Notify the firewall module to delete the old rules.
            task = QueueTask(TASK_FIREWALL_DELETE_SLOT,
                             src_module=self.get_name(),
                             dest_module=SilentDuneClientFirewallModule().get_name(),
                             data=slot)
            self.send_parent_task(task)

            # Notify the firewall module to load the new rules.
            task = QueueTask(TASK_FIREWALL_INSERT_RULES,
                             src_module=self.get_name(),
                             dest_module=SilentDuneClientFirewallModule().get_name(),
                             data=rules)
            self.send_parent_task(task)

            time.sleep(1)  # Let the firewall apply the rule changes
        else:
            _logger.info('{0}: {1}: discovery service did not return any rules.'.format(
                self.get_name(), class_name))

            _logger.debug('SLOTS: {0}: {1}'.format(Slots.ntp, slot))

            # If there were no rules discovered for NTP, open up access to all NTP servers.
            # In self._t_ntp_check_interval seconds we will check to see if any NTP servers are active.
            if slot == Slots.ntp and is_service_running('ntpd'):
                self._all_ntp_access_enabled = True
                _logger.debug('{0}: Asking Firewall Module to enable generic NTP access.'.format(self.get_name()))
                task = QueueTask(TASK_FIREWALL_ALLOW_ALL_NTP_ACCESS,
                                 src_module=self.get_name(),
                                 dest_module=SilentDuneClientFirewallModule().get_name())
                self.send_parent_task(task)

            return 0

        return len(rules)

    def flatten_rules(self, obj, level=0):
        """
        Takes nested rules and recursively un-winds them.
        :param obj: Rules object
        :param level: recursion level counter
        :return: rule list
        """

        if level == 0:
            self._flat_rules = list()

        try:

            # Test obj to see if it is really a rule
            if obj.to_json().encode('utf-8'):

                match = False

                for rule in self._flat_rules:
                    if hashlib.sha1(rule.to_json().encode('utf-8')).hexdigest() == \
                            hashlib.sha1(obj.to_json().encode('utf-8')).hexdigest():
                        match = True

                # Don't add rules that already exist in our flat list.
                if not match:
                    self._flat_rules.append(obj)

        except AttributeError:
            if obj and len(obj) > 0:
                for o in obj:
                    self.flatten_rules(o, (level+1))

        if level == 0:
            rules = self._flat_rules
            self._flat_rules = None  # clear list.
            return rules



#
# Authors: Robert Abram <robert.abram@entpack.com>,
#
# Copyright (C) 2015-2016 EntPack
# see file 'LICENSE' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import hashlib
import logging
import os
import pkg_resources
import subprocess
from subprocess import CalledProcessError
import sys

from silentdune_client import modules
from silentdune_client.builders import iptables as ipt
from silentdune_client.models.iptables_rules import IPMachineSubset, IPRulesFileWriter
from silentdune_client.modules.firewall.manager.slots import Slots
from silentdune_client.modules.firewall.manager.iptables_utils import create_iptables_tcp_ingress_egress_rule, \
    create_iptables_loopback, create_iptables_ingress_egress_forward_reject_rule, check_transport_value, \
    create_iptables_ingress_egress_reject_rule
from silentdune_client.modules.firewall.manager.utils import create_generic_dns_egress_rules, \
    create_generic_ntp_egress_rules
from silentdune_client.utils.misc import validate_network_address, is_service_running, which

# Import other module's TASK defines, set to None if module is not present.
# TODO:
try:
    from silentdune_client.modules.comm.sd_server import TASK_SEND_SERVER_ALERT
    from silentdune_client.modules.comm.sd_server import TASK_SEND_SERVER_ALERT_2  # noqa
except ImportError:
    TASK_SEND_SERVER_ALERT = None


_logger = logging.getLogger('sd-client')

""" Tell the firewall to reload all rules from disk """
TASK_FIREWALL_RELOAD_RULES = 100

""" Tell the firewall to allow all DNS queries """
TASK_FIREWALL_ALLOW_ALL_DNS_ACCESS = 110
TASK_FIREWALL_DISABLE_ALL_DNS_ACCESS = 111

""" Tell the firewall to allow all HTTP(S) queries """
TASK_FIREWALL_ALLOW_ALL_HTTP_ACCESS = 120
TASK_FIREWALL_DISABLE_ALL_HTTP_ACCESS = 121

""" Tell the firewall to allow all NTP queries """
TASK_FIREWALL_ALLOW_ALL_NTP_ACCESS = 140
TASK_FIREWALL_DISABLE_ALL_NTP_ACCESS = 141

TASK_FIREWALL_INSERT_RULES = 201
TASK_FIREWALL_DELETE_RULES = 202
TASK_FIREWALL_DELETE_SLOT = 203


# Define the available Module classes.
module_list = {
    'Silent Dune Firewall Manager': {
        'module': 'SilentDuneClientFirewallModule',
    },
}


class SilentDuneClientFirewallModule(modules.BaseModule):
    """ Silent Dune Server Module """

    priority = 0  # Highest loading priority.

    _rules = None  # List with IPTablesMachineSubset objects.
    _startup = True

    _disable_auto_ssh = False
    _sshd_running = False
    _sshd_rules = None

    _allowed_networks = None  # None == All

    def __init__(self):

        self._arg_name = 'firewall'
        self._config_section_name = 'firewall'

        # Enable multi-threading
        self.wants_processing_thread = True

        try:
            self._version = pkg_resources.get_distribution(__name__).version
        except:
            self._version = 'unknown'

    def install_module(self):
        """
        Virtual Override
        """
        return True

    def add_installer_arguments(self, parser):
        """
        Virtual Override
        Add our module's argparser arguments
        """

        # Create a argument group for our module
        group = parser.add_argument_group('firewall', 'Silent Dune Firewall')

        group.add_argument(_('--allowed-networks'), help=_('Restrict access to specified networks.'),  # noqa
                           default='', type=str, required=False)
        group.add_argument(_('--disable-auto-ssh'), help=_('Disable auto SSH rule creation.'),  # noqa
                           default=False, action='store_true')  # noqa

    def validate_arguments(self, args):
        """
        Virtual Override
        Validate command line arguments and save values to our configuration object.
        :param args: An argparse object.
        """

        self._allowed_networks = args.allowed_networks

        # if we have network addresses, check them.
        if self._allowed_networks:
            if not validate_network_address(self._allowed_networks):
                print('sdc-install: argument --allowed-networks value invalid.')
                return False

        if '--allowed-networks' not in sys.argv and '--iso-all-rules' in sys.argv:
            print('sdc-install: warning, --iso-all-rules parameter has no affect.')

        self._disable_auto_ssh = args.disable_auto_ssh

        return True

    def validate_config(self, config):
        """
        Virtual Override
        Validate configuration file arguments and save values to our config object.
        :param config: A ConfigParser object.
        """

        self._allowed_networks = config.get(self._config_section_name, 'allowed_networks')

        # if we have network addresses, check them.
        if self._allowed_networks:
            if not validate_network_address(self._allowed_networks):
                _logger.error('{0} invalid Allowed Networks value.'.format(self.get_name()))
                return False

        self._disable_auto_ssh = \
            True if config.get(self._config_section_name, 'disable_auto_ssh').lower() == 'yes' else False

        return True

    def prepare_config(self, config):
        """
        Virtual Override
        Return the configuration file structure. Any new configuration items should be added here.
        Note: The order should be reverse of the expected order in the configuration file.
        """

        config.set(self._config_section_name, 'allowed_networks',
                   self._allowed_networks if self._allowed_networks else '')
        config.set(self._config_section_name, 'disable_auto_ssh', 'yes' if self._disable_auto_ssh else 'no')

        config.set_comment(self._config_section_name, self._config_section_name,
                           _('; Silent Dune Firewall Configuration\n'))  # noqa
        config.set_comment(self._config_section_name, 'allowed_networks',
                           _('; Set allowed networks this system can connect to.\n' +  # noqa
                             '; List of networks in CIDR notation, delimited with spaces.\n' +  # noqa
                             '; Empty value allows access to all networks.'))  # noqa
        config.set_comment(self._config_section_name, 'disable_auto_ssh',
                           _('; Disable the auto SSH rule generation.\n'))  # noqa

        return True

    def service_startup(self):

        _logger.debug('{0} module startup called'.format(self.get_name()))

        p = subprocess.Popen(['modprobe', 'ip_conntrack'], stdin=subprocess.PIPE)
        p.communicate()
        result = p.wait()
        if result:
            _logger.error('{0}: kernel module loading of ip_conntrack reported an error.'.format(self.get_name()))

        p = subprocess.Popen(['modprobe', 'ip_conntrack_ftp'], stdin=subprocess.PIPE)
        p.communicate()
        result = p.wait()
        if result:
            _logger.error('{0}: kernel module loading of ip_conntrack_ftp reported an error.'.format(self.get_name()))

        self.restore_iptables()
        self.load_rule_bundles()

        return True

    def service_shutdown(self):

        _logger.debug('{0} thread shutdown called'.format(self.get_name()))
        self.save_iptables()

        # Flush rules and chains from iptables and ip6tables.
        for i in ['iptables', 'ip6tables']:
            try:
                subprocess.call([i, '--flush'])
                subprocess.call([i, '--delete-chain'])
            except CalledProcessError:
                pass

    def process_loop(self):
        # _logger.debug('{0} processing loop called'.format(self.get_name()))

        # Add SSH access rules.
        if self._startup:
            # Add loopback rules
            self.add_firewall_rule(self.get_loopback_rules())

            # Add sshd service rules
            if is_service_running('sshd'):
                self._sshd_rules = self.create_ssh_rules()
                self.add_firewall_rule(self._sshd_rules)
                self._sshd_running = True

            # Add rejection rules.
            self.add_firewall_rule(self.create_reject_rules())

            # Add network isolation rules
            if self._allowed_networks:
                self.add_firewall_rule(self.create_network_isolation_rules())

            self.write_rules_to_iptables_file()
            self.restore_iptables()

            self._startup = False

        # Check to see if sshd is running or not.
        if is_service_running('sshd'):
            if not self._sshd_running:
                self._sshd_rules = self.create_ssh_rules()
                self.add_firewall_rule(self._sshd_rules)
                self._sshd_running = True
                self.write_rules_to_iptables_file()
                self.restore_iptables()
        else:
            if self._sshd_running and self._sshd_rules:
                _logger.debug('{0}: removing sshd service rules.'.format(self.get_name()))
                self.del_firewall_rule(self._sshd_rules)
                self._sshd_running = False
                self.write_rules_to_iptables_file()
                self.restore_iptables()

    def process_task(self, task):

        if task:

            t_id = task.get_task_id()

            if t_id == TASK_FIREWALL_RELOAD_RULES:
                _logger.debug('{0}: {1}: reloading firewall rules.'.format(
                    self.get_name(), 'TASK_FIREWALL_RELOAD_RULES'))
                self.restore_iptables()

            if t_id == TASK_FIREWALL_INSERT_RULES:
                _logger.debug('{0}: {1}: received rules from module {2}.'.format(
                    self.get_name(), 'TASK_FIREWALL_INSERT_RULES', task.get_sender()))
                self.add_firewall_rule(task.get_data())
                self.write_rules_to_iptables_file()
                self.restore_iptables()

            if t_id == TASK_FIREWALL_DELETE_RULES:
                _logger.debug('{0}: {1}: received rules from module {2}.'.format(
                    self.get_name(), 'TASK_FIREWALL_DELETE_RULES', task.get_sender()))
                self.del_firewall_rule(task.get_data())
                self.write_rules_to_iptables_file()
                self.restore_iptables()

            if t_id == TASK_FIREWALL_DELETE_SLOT:
                _logger.debug('{0}: {1}: received rules from module {2}.'.format(
                    self.get_name(), 'TASK_FIREWALL_DELETE_SLOT', task.get_sender()))
                self.del_firewall_slot(task.get_data())
                self.write_rules_to_iptables_file()
                self.restore_iptables()

            if t_id == TASK_FIREWALL_ALLOW_ALL_DNS_ACCESS:
                _logger.debug('{0}: {1}: request for ALL DNS access from module {2}.'.format(
                              self.get_name(), 'TASK_FIREWALL_ALLOW_ALL_DNS_ACCESS', task.get_sender()))
                rules = create_generic_dns_egress_rules()
                self.add_firewall_rule(rules)
                self.write_rules_to_iptables_file()
                self.restore_iptables()

            if t_id == TASK_FIREWALL_DISABLE_ALL_DNS_ACCESS:
                _logger.debug('{0}: {1}: request to remove ALL DNS access from module {2}.'.format(
                              self.get_name(), 'TASK_FIREWALL_ALLOW_ALL_DNS_ACCESS', task.get_sender()))
                rules = create_generic_dns_egress_rules()
                self.del_firewall_rule(rules)
                self.write_rules_to_iptables_file()
                self.restore_iptables()

            if t_id == TASK_FIREWALL_ALLOW_ALL_NTP_ACCESS:
                _logger.debug('{0}: {1}: request for all NTP access from module {2}.'.format(
                              self.get_name(), 'TASK_FIREWALL_ALLOW_ALL_NTP_ACCESS', task.get_sender()))
                rules = create_generic_ntp_egress_rules()
                self.add_firewall_rule(rules)
                self.write_rules_to_iptables_file()
                self.restore_iptables()

            if t_id == TASK_FIREWALL_DISABLE_ALL_NTP_ACCESS:
                _logger.debug('{0}: {1}: request to remove all NTP access from module {2}.'.format(
                              self.get_name(), 'TASK_FIREWALL_ALLOW_ALL_NTP_ACCESS', task.get_sender()))
                rules = create_generic_ntp_egress_rules()
                self.del_firewall_rule(rules)
                self.write_rules_to_iptables_file()
                self.restore_iptables()

    def load_rule_bundles(self):
        """
        Load the user defined json bundle files, each file is a IPMachineSubset json object.
        :return:
        """
        ol = list()
        for file in os.listdir(self.node_info.config_root):
            if '.bundle' in file:
                try:
                    with open(os.path.join(self.node_info.config_root, file)) as handle:
                        mss = IPMachineSubset(handle)

                        if 2000 <= mss.slot <= 8999:  # only load user defined slot range numbers.
                            ol.append(mss)
                except:
                    _logger.error('{0}: Bundle file is corrupt, unable to load {1}.'.format(self.get_name(), file))
                    continue

        if len(ol) > 0:
            ol.sort(key=lambda x: x.slot)

        self._rules = ol

    def restore_iptables(self):
        """
        Load the iptables save file and load it into the kernel.
        This is only called on startup.
        """
        # Load rule files into kernel
        for v in [u'ipv4', u'ipv6']:

            file = os.path.join(self.node_info.config_root, u'{0}.rules'.format(v))
            if os.path.exists(file):
                try:
                    with open(file) as handle:
                        data = handle.read()

                    if v == u'ipv4':
                        p = subprocess.Popen([self.node_info.iptables_restore, '-c'], stdin=subprocess.PIPE)
                    else:
                        p = subprocess.Popen([self.node_info.ip6tables_restore, '-c'], stdin=subprocess.PIPE)
                    p.communicate(data.encode('utf-8'))
                    result = p.wait()

                    if result:
                        _logger.error('{0}: iptables-restore reported an error loading data.'.format(self.get_name()))

                except ValueError:
                    _logger.error('{0}: Invalid arguments passed to iptables-restore.'.format(self.get_name()))
                except OSError:
                    _logger.error('{0}: Loading IPv4 rules into kernel failed.'.format(self.get_name()))
            else:
                _logger.error('{0}: Rules file ({1}) not found.'.format(self.get_name(), file))

                # TODO: The SD Server module should be notified if there is any error loading a rule file.

    def save_iptables(self):
        """
        Dump the iptables information from the kernel and save it to the restore file.
        This is only called on shutdown.
        """

        # Load rule files into kernel
        for v in [u'ipv4', u'ipv6']:

            file = os.path.join(self.node_info.config_root, u'{0}.rules'.format(v))
            try:
                if v == u'ipv4':
                    p = subprocess.Popen([self.node_info.iptables_save, '-c'], stdout=subprocess.PIPE)
                else:
                    p = subprocess.Popen([self.node_info.ip6tables_save, '-c'], stdout=subprocess.PIPE)

                data = p.communicate()[0]
                result = p.wait()

                if result:
                    _logger.error('{0}: iptables-save reported an error.'.format(self.get_name()))
                else:
                    with open(file, 'w') as handle:
                        handle.write(str(data))

            except ValueError:
                _logger.error('{0}: Invalid arguments passed to iptables-save.'.format(self.get_name()))
            except OSError:
                _logger.error('{0}: Saving IPv4 rules from kernel failed.'.format(self.get_name()))

            if TASK_SEND_SERVER_ALERT:
                pass
                # TODO: The SD Server module should be notified if there is any error saving a rule file.

        # Restore the selinux context if needed
        self.node_info.restorecon(self.node_info.config_root)

    def add_firewall_rule(self, obj):
        """
        Add a IPTablesMachineSubset to the list of rules
        :param obj: IPTablesMachineSubset object or array of IPTablesMachineSubset objects.
        :return: True if rules were successfully added, otherwise False.
        """
        try:
            nmss = hashlib.sha1(obj.to_json().encode('utf-8')).hexdigest()

            # Loop through the current rules and see if the rule already exists.
            for mss in self._rules:
                if hashlib.sha1(mss.to_json().encode('utf-8')).hexdigest() == nmss:
                    # _logger.debug('{0}: Rule has already been added to rule list.'.format(self.get_name()))
                    return True

            self._rules.append(obj)
            self._rules.sort(key=lambda x: x.slot)

            return True

        except AttributeError:

            try:
                # See if we have a list of IPTablesMachineSubset objects. If so, recursively add them.
                if len(obj) > 0:
                    for o in obj:
                        if o:
                            self.add_firewall_rule(o)
                    # for mss in self._rules:
                    #     _logger.debug('{0} Adding rule ({1}): {2}'.format(self.get_name(), mss.slot, mss.name))
                    _logger.debug('{0}: Added {1} rules.'.format(self.get_name(), len(obj)))
                    return True
            except TypeError:
                _logger.error('{0} Invalid IPMachineSubset passed, unable to add rules.'.format(self.get_name()))

        return False

    def del_firewall_rule(self, obj):
        """
        Remove a IPTablesMachineSubset from the list of rules
        :param obj:
        :return:
        """

        try:
            nmss = hashlib.sha1(obj.to_json().encode('utf-8')).hexdigest()

            # Loop through the current rules and see if the rule already exists.
            for mss in self._rules:
                if hashlib.sha1(mss.to_json().encode('utf-8')).hexdigest() == nmss:
                    self._rules.remove(mss)
                    self._rules.sort(key=lambda x: x.slot)
                    return True

        except AttributeError:

            try:
                # See if we have a list of IPTablesMachineSubset objects. If so, recursively remove them.
                if len(obj) > 0:
                    for o in obj:
                        self.del_firewall_rule(o)
                    _logger.debug('{0}: Removed {1} IPTablesMachineSubset object rules.'.format(
                        self.get_name(), len(obj)))
                    return True
            except TypeError:
                _logger.error('{0}: Invalid IPMachineSubset passed, unable to remove.'.format(self.get_name()))

        return False

    def del_firewall_slot(self, slot):
        """
        Remove all rules that match the slot id parameter
        :param slot: Slot ID
        :return: True if successful, otherwise False.
        """

        rules = list()

        try:
            # Loop through the current rules and see if rules of the chain already exists.
            for mss in self._rules:
                jsonobj = mss.to_dict()
                if jsonobj[u'slot'] == slot:
                    rules.append(mss)

            # delete rules of the chain
            self.del_firewall_rule(rules)

            return True

        except AttributeError:

            _logger.error('{0}: Unable to delete slot ({1}) rules.'.format(self.get_name(), slot))

        return False

    def write_rules_to_iptables_file(self):
        """
        Output to file in iptables format our list of IPTablesMachineSubset objects.
        :return:
        """
        for v in [u'ipv4', u'ipv6']:
            file = os.path.join(self.node_info.config_root, u'{0}.rules'.format(v))

            _logger.debug('{0}: Writting {1} rules to -> {2}'.format(self.get_name(), v, file))

            writer = IPRulesFileWriter(self._rules)
            writer.write_to_file(file, v)

            if not self.validate_rule_files(v, file):
                return False

        return True

    def validate_rule_files(self, protocol, file):
        """
        Validate multiple iptables rule save files.
        :param files: List of path+filenames to run iptables-restore --test on.
        :return:
        """

        if not self.node_info.root_user:
            _logger.warning('{0}: Unable to validate rules, not running as privileged user.'.format(self.get_name()))
            return True

        # Loop through files and test the validity of the file.
        if not os.path.exists(file):
            _logger.error('{0} {1} save file ({2}) does not exist.'.format(self.get_name(), protocol, file))
            return False

        with open(file) as handle:
            data = handle.read()

        if protocol == u'ipv4':
            p = subprocess.Popen([self.node_info.iptables_restore, '--test'], stdin=subprocess.PIPE)
        else:
            p = subprocess.Popen([self.node_info.ip6tables_restore, '--test'], stdin=subprocess.PIPE)

        p.communicate(data.encode('utf-8'))
        result = p.wait()

        if result:
            _logger.error('{0}: {1} validation failed on iptables save file: {2}'.format(
                self.get_name(), protocol, file))

            if TASK_SEND_SERVER_ALERT:
                # TODO: Notify server of error.
                pass

        return True

    def generate_emergency_rules(self):
        """
        Generate temp admin access only rules and set firewall
        :return:
        """
        pass

    def get_loopback_rules(self):
        """
        Create Loopback rules for internal access
        :return: Loopback rules
        """

        rules = list()

        _logger.debug('{0}: adding loopback interface rules.'.format(self.get_name()))
        rules.append(create_iptables_loopback(Slots.loopback, transport=ipt.TRANSPORT_IPV4))
        rules.append(create_iptables_loopback(Slots.loopback, transport=ipt.TRANSPORT_IPV6))

        return rules

    def create_ssh_rules(self):
        """
        Create SSH rules for administrative access.
        :return: SSH Rules.
        """

        rules = list()

        # See if we should not add any ssh rules.
        if self._disable_auto_ssh:
            return rules

        conf = u'/etc/ssh/sshd_config'
        port = 22

        # Check the sshd config and see if it is using a non-standard port.
        if not os.path.exists(conf):
            _logger.warning('{0}: sshd_config not found, using standard SSH port.'.format(self.get_name()))

        try:
            with open(conf) as handle:
                content = handle.readlines()

            if content:
                for line in content:
                    if line.lower().strip().startswith('port '):
                        port = int(line.split()[1])
                        break
        except:
            pass

        _logger.debug('{0}: adding sshd service rules.'.format(self.get_name()))
        rules.append(create_iptables_tcp_ingress_egress_rule(None, port, Slots.admin, transport=ipt.TRANSPORT_IPV4))
        rules.append(create_iptables_tcp_ingress_egress_rule(None, port, Slots.admin, transport=ipt.TRANSPORT_IPV6))

        return rules

    def create_reject_rules(self):
        """
        Create auto rejection rules for Input, Forward and Output
        :return:
        """

        rules = list()

        _logger.debug('{0}: adding reject rules.'.format(self.get_name()))
        rules.append(create_iptables_ingress_egress_forward_reject_rule(Slots.reject, transport=ipt.TRANSPORT_IPV4))
        rules.append(create_iptables_ingress_egress_forward_reject_rule(Slots.reject, transport=ipt.TRANSPORT_IPV6))

        return rules

    def create_network_isolation_rules(self):

        rules = list()
        transport_list = list()

        _logger.debug('{0}: adding network isolation rules.'.format(self.get_name()))

        for cidr in self._allowed_networks.split(' '):

            cidr = cidr.encode('utf-8')

            if not validate_network_address(cidr):
                _logger.error('{0} unable to validate network cidr value "{1}"'.format(self.get_name(), cidr))
                continue

            # Get IP transport version from address.
            ipaddr, bits = cidr.split('/')
            transport = check_transport_value(ipaddr, ipt.TRANSPORT_AUTO)

            # Save the transports we use so we can add the reject rules later.
            if transport not in transport_list:
                transport_list.append(transport)

            rules.append(ipt.get_machine_subset(
                "network iso",
                Slots.network_iso,
                [
                    ipt.get_chain(
                        u'filter',
                        [
                            ipt.get_ring(
                                u'input',
                                transport,
                                [
                                    ipt.get_rule(source_address=cidr, jump=ipt.get_jump(target=u'RETURN'))
                                ]
                            ),
                            ipt.get_ring(
                                u'output',
                                transport,
                                [
                                    ipt.get_rule(dest_address=cidr, jump=ipt.get_jump(target=u'RETURN'))
                                ]
                            ),
                        ]
                    )
                ]
            ))

        # Create network isolation rejection rules.
        for transport in transport_list:
            rules.append(create_iptables_ingress_egress_reject_rule(Slots.network_iso, transport, 'iso reject'))

        return rules



#
# Authors: Robert Abram <robert.abram@entpack.com>
#
# Copyright (C) 2015-2016 EntPack
# see file 'LICENSE' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from datetime import datetime
from dateutil.tz import tzlocal
import logging
import re

from silentdune_client.utils.node_info import NodeInformation
from silentdune_client.models.json_base import JsonObject

_logger = logging.getLogger('sd-client')


class IPLogEvent(JsonObject):
    """
    Kernel Firewall Log Event record
    based off iptables log format information found here:
        http://search.cpan.org/~paulv/POE-Filter-Log-IPTables-0.02/IPTables.pm
    """
    event = None  # Unmodified log event string
    event_flag = None  # Our log event prefix value, usually set by the server module.  Must start with 'SDC'.

    # Top level event structure
    event_dt = None  # Timestamp of packet
    tz_offset = None  # Timezone offset from GMT in "-+00:00" format.
    iface_in = None  # Interface In - Packet received on.
    iface_out = None  # Interface Out - Packet sent on.
    src_addr = None  # Source IP Address
    dest_addr = None  # Destination IP address
    length = None  # Packet length
    tos = None  # Type of Service
    prec = None  # Precedence
    ttl = None  # Packet Time to Live value.
    packet_id = None  # Packet ID
    frag_flags = list()  # Can have "CE" (congestion), "DF" (don't fragment), or "MF" (more fragments are coming).
    protocol = None  # 'tcp', 'udp', 'icmp', or a number corresponding to the protocol in /etc/protocols.

    # Event values based on protocol value.
    # TCP and UDP
    src_port = None  # Source port
    dest_port = None  # Destination port

    # TCP only
    window = None  # Length of the tcp window.
    reserved = None  # Reserved bits.
    urgp = None  # The urgent pointer.
    tcp_flags = list()  # "CWR" (Congestion Window Reduced), "ECE" (Explicit Congestion Notification Echo),
    # "URG" (Urgent), "ACK" (Acknowledgement), "PSH" (Push), "RST" (Reset), "SYN" (Synchronize), or "FIN" (Finished)

    # UDP only
    udp_length = None  # length of the UDP packet.

    # ICMP
    icmp_type = None  # For ICMP this is a numeric value,
    icmp_code = None  # Numeric code of the ICMP packet.
    icmp_id = None  # ID of the ICMP echo packet.
    icmp_seq = None  # Sequence of the ICMP packet.
    icmp_error_header = None  # Some types of ICMP - 3 (destination unreachable), 4 (source quench), and
    # 11 (time exceeded) contain the IP and protocol headers that generated the ICMP packet.
    # This will be a IPLogEvent object.

    parent_id = None  # Points to the parent record when there is an ICMP child data record.

    leftover = ''  # Left over log event text that was not used during parsing.

    def __init__(self, event):
        self.event = event
        self._parse()

    def __str__(self):
        return self.event

    def _parse(self):

        if NodeInformation.platform == 'linux':
            return self._parse_linux()

        raise ValueError('IPLogEvent Parser: os platform not supported.')

    def _parse_linux(self):

        p = self.event

        # Look for embedded packet event log first.
        if '[' in p:
            try:
                imbed_p = p[p.index('['):p.rindex(']') + 1]
                p = p.replace(imbed_p, '')
                imbed_p = imbed_p[1:-1].strip()  # Remove braces
                self.icmp_error_header = IPLogEvent(imbed_p)  # Recursively parse this packet
            except ValueError:
                raise
                pass

        # Retrieve the timestamp and remove it from the event data.
        if 'kernel:' in p:
            # Get the timestamp
            ts = p[:p[:p.index('kernel:')].strip().rindex(' ')]
            p = p.replace(ts, '')
            self.event_dt = datetime.strptime(ts, '%b %d %H:%M:%S').replace(year=datetime.now().year,
                                                                            tzinfo=tzlocal()).isoformat()
            offset = re.search('(Z|(\+|-)[0-9][0-9]:[0-9][0-9])$', self.event_dt)
            if offset:
                self.tz_offset = offset.group(0)

            # Remove the system name and the 'kernel:' text.
            p = p[p.index(':') + 1:].strip()

            # Capture the log record type flag.
            self.event_flag = p[:p.index(':')]
            p = p.replace(self.event_flag + ':', '').strip()

        # parse event log key value pairs.
        p, self.iface_in = self._parse_event_key_value(p, 'IN')
        p, self.iface_out = self._parse_event_key_value(p, 'OUT')
        p, self.src_addr = self._parse_event_key_value(p, 'SRC')
        p, self.dest_addr = self._parse_event_key_value(p, 'DST')
        p, self.length = self._parse_event_key_value(p, 'LEN')
        p, self.tos = self._parse_event_key_value(p, 'TOS')
        p, self.prec = self._parse_event_key_value(p, 'PREC')
        p, self.ttl = self._parse_event_key_value(p, 'TTL')
        p, self.packet_id = self._parse_event_key_value(p, 'ID')
        p, self.protocol = self._parse_event_key_value(p, 'PROTO')

        if self.protocol == 'TCP':
            p, self.window = self._parse_event_key_value(p, 'WINDOW')
            p, self.reserved = self._parse_event_key_value(p, 'RES')
            p, self.urgp = self._parse_event_key_value(p, 'URGP')
            p, self.src_port = self._parse_event_key_value(p, 'SPT')
            p, self.dest_port = self._parse_event_key_value(p, 'DPT')

        if self.protocol == 'UDP':
            p, self.udp_length = self._parse_event_key_value(p, 'LEN')
            p, self.src_port = self._parse_event_key_value(p, 'SPT')
            p, self.dest_port = self._parse_event_key_value(p, 'DPT')

        if self.protocol == 'ICMP':
            p, self.icmp_code = self._parse_event_key_value(p, 'CODE')
            p, self.icmp_seq = self._parse_event_key_value(p, 'SEQ')
            p, self.icmp_type = self._parse_event_key_value(p, 'TYPE')
            p, self.icmp_id = self._parse_event_key_value(p, 'ID')

        if self.protocol == 'TCP':
            p, self.tcp_flags = self._parse_event_flags(p, self.tcp_flags, 'CWR')
            p, self.tcp_flags = self._parse_event_flags(p, self.tcp_flags, 'ECE')
            p, self.tcp_flags = self._parse_event_flags(p, self.tcp_flags, 'URG')
            p, self.tcp_flags = self._parse_event_flags(p, self.tcp_flags, 'ACK')
            p, self.tcp_flags = self._parse_event_flags(p, self.tcp_flags, 'PSH')
            p, self.tcp_flags = self._parse_event_flags(p, self.tcp_flags, 'RST')
            p, self.tcp_flags = self._parse_event_flags(p, self.tcp_flags, 'SYN')
            p, self.tcp_flags = self._parse_event_flags(p, self.tcp_flags, 'FIN')

        # parse event log fragment flags
        p, self.frag_flags = self._parse_event_flags(p, self.frag_flags, 'CE')
        p, self.frag_flags = self._parse_event_flags(p, self.frag_flags, 'DF')
        p, self.frag_flags = self._parse_event_flags(p, self.frag_flags, 'MF')

        # Save any unhandled event text.
        self.leftover = p.strip()

    def _parse_event_key_value(self, p, key):
        """
        Return the Key value and remove the Key Value from the string parameter.
        Only looks at and removes the first occurrence of a Key.  Some key names are duplicated in an event.
        """
        key += '='
        p += ' '  # Add a space at the end in case the key is the last element.
        try:
            s = p[p.index(key):p[p.index(key):].index(' ') + p.index(key)]
            v = s.split('=')[1]
            return p.replace(s, '').replace('  ', ' ').strip(), v if v else None
        except ValueError:
            if key == 'SRC':
                raise
            pass

        return p, None

    def _parse_event_flags(self, p, l, flag):
        """
        Look for the flag in the event data, remove if found and add to list object.
        :param p: Event data
        :param l: List object
        :param flag: Flag to search for.
        :return: Event data, list
        """
        if flag in p:
            l.append(flag)
            p = p.replace(flag, '').replace('  ', ' ').strip()

        return p, l



#
# Authors: Robert Abram <robert.abram@entpack.com>
#
# Copyright (C) 2015-2016 EntPack
# see file 'LICENSE' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging

from silentdune_client.utils.node_info import NodeInformation
from silentdune_client.builders import iptables as ipt
from silentdune_client.utils.misc import is_valid_ipv4_address, is_valid_ipv6_address

# !!! Slot list from: SilentDune/silentdune_server/apps/rules/models.py !!!
# RULE_MACHINE_SUBSET_SLOTS = (
#     (110, _('Administration')),
#     # (120, _('Silent Dune Server')),
#     (130, _('Dynamic Name Service')),
#     (140, _('Network Time Protocol')),
#     (150, _('System Update')),
#     (160, _('Identity Services')),
#     (170, _('Dynamic Host Configuration Protocol')),
#     #(1000, _('Node Custom Rules')),
#     (2000, _('User Defined')),  # 2000 -> 8999 are for user defined rules.
#     #(9000, _('Reserved')),
#     (9800, _('Logging Rules')),
#     (9900, _('Rejection Rules'))
# )

_logger = logging.getLogger('sd-client')


class BaseDiscoveryService(object):
    """
    Simple base object for auto service discovery.
    """

    _slot = 0  # Set this to the proper slot level value.
    _config_property_name = 'unknown'  # Config property name in SilentDuneAutoServiceDiscoveryModule.
    config = None
    _parent = None

    def __init__(self, config):
        self.config = config

    def get_name(self):
        """
        :return: module name
        """
        # return self._name
        return type(self).__name__

    def get_slot(self):
        return self._slot

    def get_config_property_name(self):
        return self._config_property_name

    def discover(self, parent):
        """
        Please DO NOT override this method. Override the platform specific discover methods.
        """
        self._parent = parent  # So we can send rules to the firewall manager module directly.

        if NodeInformation.firewall_platform == 'iptables':
            return self._discover_iptables(), self._slot

    def _discover_iptables(self):
        """
        Override this method to discover external services and return either a single machine_subset object or a
        list of machine_subset objects to add to the firewall rules.

        See: modules/firewall/manage/iptables_utils.py for iptables rule builder methods.

        :return: Single iptables machine_subset object or list of objects.
        """

        pass


#
# Authors: Robert Abram <robert.abram@entpack.com>
#
# Original author is Sander Marechal <s.marechal@jejik.com>
# http://www.jejik.com/articles/2007/02/a_simple_unix_linux_daemon_in_python/
#
# Copyright (C) 2015-2016 EntPack
# see file 'LICENSE' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# Changes: Robert Abram
#
# Added code to demote daemon process to specified uid and gid
# Added debug logging
# Re-coded for python 3.x


import atexit
import grp
import logging
import os
import pwd
import sys
import time
from signal import SIGTERM

_logger = logging.getLogger('sd-client')


class Daemon(object):
    """
    A generic daemon class.

    Usage: subclass the Daemon class and override the run() method
    """

    def __init__(self, procbase=None, dirmask=None, pidfile=None, uid='nobody', gid='nobody',
                 stdin='/dev/null', stdout='/dev/null', stderr='/dev/null'):
        self.procbase = procbase
        self.dirmask = dirmask

        if procbase and pidfile:
            self.pidfile = os.path.join(procbase, pidfile)

        self.stdin = stdin
        self.stdout = stdout
        self.stderr = stderr
        self.uid = uid
        self.gid = gid

    def daemonize(self):
        """
        do the UNIX double-fork magic, see Stevens' "Advanced
        Programming in the UNIX Environment" for details (ISBN 0201563177)
        http://www.erlenstar.demon.co.uk/unix/faq_2.html#SEC16
        """

        # Lookup group and user id if we are the root user
        if self.uid != 'nobody' and os.getuid() == 0:
            try:
                groupinfo = grp.getgrnam(self.gid)
                _logger.debug('our group info. n: {0}, i:{1}'.format(groupinfo.gr_name, groupinfo.gr_gid))

            except KeyError:
                _logger.critical('get daemon group id failed')
                sys.exit(1)

            try:
                userinfo = pwd.getpwnam(self.uid)
                _logger.debug('our user info n: {0}, i:{1}'.format(userinfo.pw_name, userinfo.pw_uid))
            except KeyError:
                _logger.critical('get daemon user id failed')
                sys.exit(1)

        try:
            pid = os.fork()
            if pid > 0:
                # exit first parent
                sys.exit(0)
        except os.error as err:
            _logger.critical('fork #1 of double fork failed. ({0}): {1}'.format(err.errno, err.strerror))
            sys.exit(1)

        # decouple from parent environment
        os.chdir("/")
        os.setsid()
        os.umask(0)

        # do second fork
        try:
            pid = os.fork()
            if pid > 0:
                # exit from second parent
                sys.exit(0)
        except os.error as err:
            _logger.critical('fork #2 of double fork failed. ({0}): {1}'.format(err.errno, err.strerror))
            sys.exit(1)

        # redirect standard file descriptors
        sys.stdout.flush()
        sys.stderr.flush()
        si = open(self.stdin, 'rb')
        so = open(self.stdout, 'a+b')
        se = open(self.stderr, 'a+b')
        os.dup2(si.fileno(), sys.stdin.fileno())
        os.dup2(so.fileno(), sys.stdout.fileno())
        os.dup2(se.fileno(), sys.stderr.fileno())

        # Setup proc base directory and mask
        try:
            os.makedirs(self.procbase, int(self.dirmask, 8))
        except OSError:
            pass

        # write pidfile
        atexit.register(self.delpid)
        pid = str(os.getpid())
        # open(self.pidfile, 'w+').write("%s\n" % pid)
        with open(self.pidfile, 'w+') as handle:
            handle.write("%s\n" % pid)
        os.chmod(self.pidfile, 0o440)

        # If uid != nobody try to demote the process
        if self.uid != 'nobody':
            # Assume that we can only demote the process if we are running as root
            if os.getuid() == 0:
                # Make the procbase directory and pid file are owned by our process user
                os.chown(self.procbase, userinfo.pw_uid, groupinfo.gr_gid)
                os.chown(self.pidfile, userinfo.pw_uid, groupinfo.gr_gid)
                # demote process to 'pki' user and 'secure' group
                os.setgid(groupinfo.gr_gid)
                os.setuid(userinfo.pw_uid)
            else:
                _logger.warning('not running as root, unable to demote process')

    def delpid(self):
        os.remove(self.pidfile)

    def start(self):
        """
        Start the daemon
        """

        # Check for a pidfile to see if the daemon already runs
        pid = None

        if os.path.isfile(self.pidfile):
            try:
                pf = open(self.pidfile, 'r')
                pid = int(pf.read().strip())
                pf.close()
            except os.error:
                pid = None

        if pid:
            _logger.warning('PID file already exists. daemon already running?'.format(self.pidfile))
            self.stop()

        # Start the daemon
        self.daemonize()
        self.run()

    def stop(self):
        """
        Stop the daemon
        """
        # Get the pid from the pidfile
        try:
            pf = open(self.pidfile, 'r')
            pid = int(pf.read().strip())
            pf.close()
        except os.error as err:
            _logger.debug('unknown error when attempting to get pid from pid file. ({0}): {1}'
                          .format(err.errno, err.strerror))
            pid = None

        if not pid:
            _logger.warning('pidfile {0} does not exist. daemon not running?'.format(self.pidfile))
            return  # not an error in a restart

        # Try killing the daemon process
        try:
            while 1:
                os.kill(pid, SIGTERM)
                time.sleep(0.2)
        except os.error as err:
            if err.errno == 3:  # No such process
                if os.path.exists(self.pidfile):
                    os.remove(self.pidfile)
            else:
                _logger.error('unknown error when attempting to kill process. ({0}): {1}'
                              .format(err.errno, err.strerror))
                sys.exit(1)

        try:
            os.rmdir(self.procbase)
        except OSError:
            pass

    def restart(self):
        """
        Restart the daemon
        """
        self.stop()
        self.start()

    def run(self):
        """
        You should override this method when you subclass Daemon. It will be called after the process has been
        daemonized by start() or restart().
        """



#
# Authors: Ma He <mahe.itsec@gmail.com>
#          Robert Abram <robert.abram@entpack.com>

# Copyright (C) 2015-2016 EntPack
# see file 'LICENSE' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from datetime import datetime
import fnmatch
import glob
import logging
import os
import platform
import shlex
import socket
import subprocess
import time

try:
    from urllib import parse
except ImportError:
    # < 3.0
    import urlparse as parse

import ConfigParser
from silentdune_client.modules import QueueTask
from silentdune_client.modules.firewall.manager import SilentDuneClientFirewallModule, \
    TASK_FIREWALL_INSERT_RULES
from silentdune_client.builders import iptables as ipt
from silentdune_client.modules.automation.auto_discovery.base_discovery_service import BaseDiscoveryService
from silentdune_client.modules.firewall.manager.iptables_utils import create_iptables_egress_ingress_rule, \
    create_iptables_egress_rule_dest, create_iptables_ingress_rule_source
from silentdune_client.modules.firewall.manager.slots import Slots
from silentdune_client.utils.misc import which

_logger = logging.getLogger('sd-client')


class SystemUpdatesDiscovery(BaseDiscoveryService):
    """
    Auto discover SystemUpdates like apt-get and yum.
    """

    _rebuild_rules_interval = 5  # Rebuild the updates cache rules every 5 days.

    _slot = Slots.updates
    _config_section_name = 'auto_discovery'
    _config_property_name = '_disable_auto_updates'
    _dist = ''
    _dist_version = ''
    _machine = ''
    _disable_auto_updates_ftp = None
    _disable_auto_updates_rsync = None

    _rebuild_rules = True  # Rebuild the rules
    _rebuild_cache = False  # Should we rebuild the package manager cache?
    _cache_last_rebuilt = None  # Time stamp last time cache was rebuilt.

    _hostnames = list()  # Previously added hostnames

    _repo_manager = None
    _repo_cache_base = ''
    _repo_config_base = ''
    _repo_service_base = ''  # openSUSE/SLES systems

    _saved_rules = None

    def __init__(self, config):
        super(SystemUpdatesDiscovery, self).__init__(config)
        self._disable_auto_updates_ftp = True if self.config.get(
            self._config_section_name, 'disable_auto_updates_ftp').lower() == 'yes' else False

    def _discover_iptables(self):
        """
        Virtual Override
        :return: Firewall Rules
        """

        rules = list()

        # Find the system package manager executable and setup our environment info
        if not self._repo_manager and not self.discover_pkg_manager():
            _logger.error('{0}: unable to continue setting up rules for updates.'.format(self.get_name()))
            return None

        # Test to see if DNS lookup is available right now.  If not, just return and wait until DNS is working.
        if not self.resolve_hostname('example.org', 80):
            return None

        # Every 5 days rebuild the rules.
        if self._cache_last_rebuilt:
            days = int((datetime.now() - self._cache_last_rebuilt).days)
            if days >= self._rebuild_rules_interval:
                self._rebuild_rules = True

        rules.append(self.iptables_updates())

        return rules

    def discover_pkg_manager(self):
        """
        Find the system package manager executable
        :return: True if found, otherwise False
        """

        self._dist = platform.dist()[0].lower()
        self._dist_version = platform.dist()[1]
        self._dist_version = self._dist_version.split('.')[0]
        self._machine = platform.machine()

        if self._dist in 'ubuntu debian':
            self._repo_manager = which('apt-get')
            # Nothing else to do.

        elif self._dist in 'centos redhat fedora':

            self._repo_config_base = '/etc/yum.repos.d/*.repo'

            self._repo_manager = which('dnf')
            self._repo_cache_base = '/var/cache/dnf'

            if not self._repo_manager:
                self._repo_manager = which('yum')
                self._repo_cache_base = '/var/cache/yum/{0}/{1}'.format(self._machine, self._dist_version)

        elif self._dist in 'suse':
            self._repo_manager = which('zypper')
            self._repo_config_base = '/etc/zypp/repos.d/*.repo'
            self._repo_service_base = '/etc/zypp/services.d/*.service'
            # No metalink cache until suse implements metalinks in zypper

        else:
            _logger.error('{0}: unsupported distribution ({1})'.format(self.get_name(), self._dist))
            return False

        if not self._repo_manager:
            _logger.error('{0}: unable to find package manager executable for {1}'.format(self.get_name(), self._dist))
            return False

        return True

    def iptables_updates(self):

        rules = list()

        if self._dist in 'ubuntu debian':
            rules.append(self.iptables_updates_apt())

        elif self._dist in 'centos redhat fedora':
            rules.append(self.add_repository_rules())

        elif self._dist in 'suse':
            rules.append(self.add_repository_rules())

        return rules

    def iptables_updates_apt(self):

        rules = list()

        if not os.path.exists('/etc/apt/sources.list'):
            _logger.error('{0}: /etc/apt/sources.list not found.'.format(self.get_name()))
            return None

        # Get all nameserver ip address values
        with open('/etc/apt/sources.list') as handle:

            for line in handle:

                if line.startswith('deb '):
                    url = line.split()[1]
                    if url:
                        rules.append(self.add_rule_by_url(url))

        return rules

    def add_repository_rules(self):
        """
        Add repository rules for rpm based systems.
        """

        # If it is not time to rebuild the package manager cache, just return the rules we have.
        if not self._rebuild_rules:
            return self._saved_rules

        self._rebuild_cache = False

        # reset hostnames list
        self._hostnames = list()

        rules = list()
        base_urls = list()
        mirror_urls = list()

        _logger.debug('{0}: adding rules for {1} repositories'.format(self.get_name(), self._dist))

        # Loop through all the repo files and gather url information
        repofiles = glob.glob(self._repo_config_base)

        # Add in any zypper service files.
        if self._dist in 'suse':
            repofiles += glob.glob(self._repo_service_base)

        for repofile in repofiles:

            config = ConfigParser.ConfigParser()
            if config.read(repofile):

                sections = config.sections()

                # Loop through sections looking for enabled repositories.
                for section in sections:

                    if config.has_option(section, 'enabled'):
                        enabled = config.getint(section, 'enabled')
                    else:
                        enabled = 1

                    if not enabled:
                        continue

                    _logger.debug('{0}: adding urls for section: {1}'.format(self.get_name(), section))

                    url = None

                    if config.has_option(section, 'metalink'):
                        url = config.get(section, 'metalink')
                        self._rebuild_cache = True
                        if url:
                            mirror_urls.append([section, url])

                    elif config.has_option(section, 'mirrorlist'):
                        url = config.get(section, 'mirrorlist')
                        self._rebuild_cache = True
                        if url:
                            mirror_urls.append([section, url])

                    elif config.has_option(section, 'baseurl'):
                        url = config.get(section, 'baseurl')
                        if url:
                            base_urls.append([section, url])

                    # Handle zypper service files.
                    elif config.has_option(section, 'url'):
                        url = config.get(section, 'url')
                        if url:
                            base_urls.append([section, url])

                    if not url:
                        _logger.debug('{0}: could not find repo section ({1}) url?'.format(self.get_name(), section))

        # Loop through all the urls and add rules for them.
        for section, url in base_urls:

            # TODO: Add support for mirrorbrain style mirrorlists.

            rules.append(self.add_rule_by_url(url))

        # If we don't need to rebuild the package manager cache, just return the rules we have.
        if not self._rebuild_cache:
            return rules

        for section, url in mirror_urls:
            rules.append(self.add_rule_by_url(url))

        # Rebuild the package manager cache
        # Open up all port 80 and port 443 connections so cache rebuild succeeds.
        all_access = list()
        all_access.append(create_iptables_egress_ingress_rule(
            '', 80, 'tcp', self._slot, transport=ipt.TRANSPORT_IPV4))
        all_access.append(create_iptables_egress_ingress_rule(
            '', 80, 'tcp', self._slot, transport=ipt.TRANSPORT_IPV6))
        all_access.append(create_iptables_egress_ingress_rule(
            '', 443, 'tcp', self._slot, transport=ipt.TRANSPORT_IPV4))
        all_access.append(create_iptables_egress_ingress_rule(
            '', 443, 'tcp', self._slot, transport=ipt.TRANSPORT_IPV6))

        # our parent will take care of clearing these rules once we return the real rules.
        self.add_url_rule_to_firewall(all_access)

        time.sleep(2)  # Give the firewall manager time to add the rules.

        if not self.rebuild_package_manager_cache():
            return rules

        # Check to see if we know where the package manage cache data is.
        if not self._repo_cache_base:
            return rules

        # loop through the mirror list and parse the mirrorlist or metalink file.
        for section, url in mirror_urls:

            file, file_type = self.get_cache_file(section)

            if file:
                if file_type == 'mirrorlist':
                    urls = self.get_mirrorlist_urls_from_file(file, section)
                else:
                    urls = self.get_metalink_urls_from_file(file, section)

                if urls:
                    for url in urls:
                        if url:
                            rules.append(self.add_rule_by_url(url))

        self._cache_last_rebuilt = datetime.now()
        self._saved_rules = rules

        return rules

    def get_cache_file(self, section):
        """
        Auto detect the mirror list path, file and type.
        :param section: config section name
        :return: file and file type
        """

        if 'yum' in self._repo_manager:
            cachepath = '{0}/{1}'.format(self._repo_cache_base, section)

        elif 'dnf' in self._repo_manager:

            # Get all the files and directories in the cache base dir
            files = os.listdir(self._repo_cache_base)

            # Filter out only the directories we are looking for
            files = fnmatch.filter(files, '{0}-????????????????'.format(section))

            # Now figure out which path is the newest one.
            cachepath = '{0}/{1}/'.format(self._repo_cache_base, max(files, key=os.path.getmtime))

        else:
            return None, ''

        if not os.path.isdir(cachepath):
            _logger.error('{0}: calculated cache path is invalid ({1})'.format(self.get_name(), cachepath))
            return None, ''

        # detect url cache file and type
        if os.path.isfile('{0}/mirrorlist.txt'.format(cachepath)):
            return '{0}/mirrorlist.txt'.format(cachepath), 'mirrorlist'

        if os.path.isfile('{0}/mirrorlist'.format(cachepath)):
            return '{0}/mirrorlist'.format(cachepath), 'mirrorlist'

        if os.path.isfile('{0}/metalink.xml'.format(cachepath)):
            return '{0}/metalink.xml'.format(cachepath), 'metalink'

        _logger.error('{0}: cache file is not found ({1})'.format(self.get_name(), cachepath))

        return None, ''

    def rebuild_package_manager_cache(self):
        """
        Have the package manager clean and rebuild it's cache information
        :return: True if successful, otherwise False
        """

        if self._dist in 'suse':
            cmd_clean = '{0} clean'.format(self._repo_manager)
            cmd_make = '{0} refresh'.format(self._repo_manager)

        elif self._dist in 'centos redhat fedora':
            cmd_clean = '{0} clean metadata'.format(self._repo_manager)
            cmd_make = '{0} makecache fast'.format(self._repo_manager)

        else:
            _logger.error('{0}: unsupported package manager, unable to rebuild cache.')
            return False

        _logger.debug('{0}: rebuilding {1} package manager cache data'.format(self.get_name(), self._dist))

        # Clean package manager cache
        p = subprocess.Popen(shlex.split(cmd_clean), stdout=subprocess.PIPE)
        stdoutdata, stderrdata = p.communicate()
        p.wait()

        if stderrdata:
            _logger.error('{0}: cleaning package manager cache failed.'.format(self.get_name()))
            return False

        # Rebuild the package manager cache
        p = subprocess.Popen(shlex.split(cmd_make), stdout=subprocess.PIPE)
        stdoutdata, stderrdata = p.communicate()
        p.wait()

        if stderrdata:
            _logger.error('{0}: rebuilding package manager cache failed.'.format(self.get_name()))
            return False

        return True

    def add_url_rule_to_firewall(self, rules):
        """
        Add rules allowing access immediately to the firewall.
        :param rules: rules list
        """
        if not rules:
            return

        # Notify the firewall module to reload the rules.
        task = QueueTask(TASK_FIREWALL_INSERT_RULES,
                         src_module=self._parent.get_name(),
                         dest_module=SilentDuneClientFirewallModule().get_name(),
                         data=rules)
        self._parent.send_parent_task(task)

        time.sleep(2)  # Give the firewall manager time to add the rule to the kernel

    def resolve_hostname(self, hostname, port):
        """
        Return a single or multiple IP addresses from the hostname parameter
        :param hostname: hostname string, IE: www.example.org
        :param port:
        :return: list of IP addresses
        """

        ipaddrs = list()

        if hostname:
            if port:
                if 1 < port <= 65536:
                    try:
                        ais = socket.getaddrinfo(hostname, port, 0, 0, socket.IPPROTO_TCP)
                    except:
                        _logger.debug('{0}: error resolving host: {1}:{2}'.format(self.get_name(), hostname, port))
                        return None

                    for result in ais:
                        ipaddr = result[-1][0]
                        ipaddrs.append(ipaddr)

        return ipaddrs

    def add_rule_by_url(self, url):
        """
        :param url: Complete url string
        :return: urlparse URI
        """

        rules = list()

        if not url:
            _logger.debug('{0}: empty url given, unable to add rule'.format(self.get_name()))
            return None

        try:
            uri = parse.urlparse(url)
        except:
            _logger.error('{0}: error parsing url ({1})'.format(self.get_name(), url))
            return None

        if not uri.scheme or not uri.hostname or uri.scheme not in ['http', 'https', 'ftp', 'rsync']:
            return None

        # If this is an FTP url...
        if uri.scheme == 'ftp':
            return self.add_ftp_rule_by_url(uri)

        port = uri.port

        if not port:

            if uri.scheme == 'http':
                port = 80
            elif uri.scheme == 'https':
                port = 443
            elif uri.scheme == 'rsync':
                port = 873

        key = uri.hostname + ':' + str(port)

        # Avoid duplicate urls.
        if key not in self._hostnames:
            self._hostnames.append(key)

            ipaddrs = self.resolve_hostname(uri.hostname, port)

            if ipaddrs:
                for ipaddr in ipaddrs:

                    # _logger.debug('{0}: adding ip: {1} from hostname: {2}'.format(
                    #     self.get_name(), uri.scheme + '://' + ipaddr, uri.hostname))

                    rules.append(create_iptables_egress_ingress_rule(
                        ipaddr, port, 'tcp', self._slot, transport=ipt.TRANSPORT_AUTO))

                    _logger.debug('{0}: host: {1}  ip: {2}:{3}'.format(self.get_name(), uri.hostname, ipaddr, port))

            return rules

        return None

    def add_ftp_rule_by_url(self, uri):
        """
        Add rules to allow FTP access based on uri value
        :param uri: urlparse uri value
        :return: rules
        """

        # Check to make sure we can add ftp rules.
        if self._disable_auto_updates_ftp:
            return None

        rules = list()

        ipaddrs = self.resolve_hostname(uri.hostname, 21)

        if ipaddrs:
            for ipaddr in ipaddrs:

                _logger.debug('{0}: adding ip: {1} from hostname: {2}'.format(
                    self.get_name(), uri.scheme + '://' + ipaddr, uri.hostname))

                # FTP control
                rules.append(create_iptables_egress_ingress_rule(ipaddr, 21, 'tcp', self._slot,
                                                                 transport=ipt.TRANSPORT_AUTO))
                # FTP data transfer
                rules.append(create_iptables_egress_rule_dest(ipaddr, 20, 'tcp', self._slot, 'ESTABLISHED',
                                                              transport=ipt.TRANSPORT_AUTO))
                rules.append(
                    create_iptables_ingress_rule_source(ipaddr, 20, 'tcp', self._slot, 'ESTABLISHED,RELATED',
                                                        transport=ipt.TRANSPORT_AUTO))
                rules.append(
                    create_iptables_egress_rule_dest(ipaddr, None, 'tcp', self._slot, 'ESTABLISHED,RELATED',
                                                     transport=ipt.TRANSPORT_AUTO))
                rules.append(
                    create_iptables_ingress_rule_source(ipaddr, None, 'tcp', self._slot, 'ESTABLISHED',
                                                        transport=ipt.TRANSPORT_AUTO))

        return rules

    def get_mirrorlist_urls_from_file(self, file, section):

        urls = list()

        if not file or not os.path.isfile(file):
            _logger.debug('{0}: unable to locate mirrorlist file ({1})'.format(self.get_name(), file))
            return urls

        with open(file) as handle:
            lines = handle.read().split('\n')

            if lines:
                for line in lines:
                    if line.strip():
                        urls.append(line.strip())

        _logger.debug('{0}: retrieving mirrorlist urls for "{1}" succeeded'.format(self.get_name(), section))
        return urls

    def get_metalink_urls_from_file(self, file, section):

        urls = list()

        if not file or not os.path.isfile(file):
            _logger.debug('{0}: unable to locate metalink file ({1})'.format(self.get_name(), file))
            return urls

        with open(file) as handle:
            lines = handle.read().split('\n')

            if lines:
                for line in lines:
                    if line.find('<url protocol=') != -1:
                        url = line.split(' >')[1].split('</url>')[0].strip()
                        if url:
                            urls.append(url)
                    if line.find('xmlns="') != -1:
                        url = line.split('xmlns="')[1].split('"')[0].strip()
                        if url:
                            urls.append(url)
                    if line.find('xmlns:mm0="') != -1:
                        url = line.split('xmlns:mm0="')[1].split('"')[0].strip()
                        if url:
                            urls.append(url)

            else:
                _logger.debug('{0}: no data read from metalink file ({1})'.format(self.get_name(), file))
                return None

        _logger.debug('{0}: retrieving metalink urls for "{1}" succeeded'.format(self.get_name(), section))

        return urls

    # def grab_mirror_list_urls(self, mirrorlist, section):
    #     """
    #     Attempt to download repository mirrorlist data using curl and return the urls.
    #     :param mirrorlist:
    #     :return: url list
    #     """
    #     urls = list()
    #
    #     if mirrorlist:
    #
    #         mirrorlist = mirrorlist.replace('$releasever', self._dist_version)
    #         mirrorlist = mirrorlist.replace('$basearch', self._machine)
    #         mirrorlist = mirrorlist.replace('$infra', 'stock')
    #
    #         c = pycurl.Curl()
    #         transport = c.IPRESOLVE_V4
    #
    #         # Force curl IPv4 DNS lookup, if that fails try IPv6 DNS lookup.
    #         while True:
    #             try:
    #                 storage = StringIO()
    #                 c = pycurl.Curl()
    #                 c.setopt(c.URL, mirrorlist.encode('utf-8'))
    #                 c.setopt(c.IPRESOLVE, transport)
    #                 c.setopt(c.WRITEFUNCTION, storage.write)
    #                 c.perform()
    #                 c.close()
    #                 content = storage.getvalue()
    #                 urllist = content.split('\n')
    #                 for url in urllist:
    #                     if url.strip():
    #                         urls.append(url.strip())
    #                 break
    #             except pycurl.error as e:
    #
    #                 if transport == c.IPRESOLVE_V4:
    #                     transport = c.IPRESOLVE_V6
    #                     continue
    #
    #                 _logger.debug('{0}: {1}: retrieving mirror list failed ({2})'.format(self.get_name(), section, e))
    #                 return None
    #
    #     _logger.debug('{0}: retrieving mirror list for section "{1}" succeeded'.format(self.get_name(), section))
    #     return urls

        # def grab_metalink_urls(self, metalink, section):
    #     """
    #     Attempt to download repository metalink data using curl and return the urls.
    #     :param metalink: metalink url
    #     :return: url list
    #     """
    #
    #     urls = list()
    #     content = None
    #
    #     if metalink:
    #         metalink = metalink.replace('$releasever', self._dist_version)
    #         metalink = metalink.replace('$basearch', self._machine)
    #
    #         c = pycurl.Curl()
    #         transport = c.IPRESOLVE_V4
    #
    #         # Force curl IPv4 DNS lookup, if that fails try IPv6 DNS lookup.
    #         while True:
    #             try:
    #                 storage = StringIO()
    #                 c = pycurl.Curl()
    #                 c.setopt(c.URL, metalink.encode('utf-8'))
    #                 c.setopt(c.IPRESOLVE, transport)
    #                 c.setopt(c.WRITEFUNCTION, storage.write)
    #                 c.perform()
    #                 c.close()
    #                 content = storage.getvalue()
    #                 break
    #             except pycurl.error as e:
    #
    #                 if transport == c.IPRESOLVE_V4:
    #                     transport = c.IPRESOLVE_V6
    #                     continue
    #
    #                 _logger.debug('{0}: retrieving metalink list failed ({1})'.format(self.get_name(), e))
    #                 return None
    #
    #         if content:
    #             lines = content.split('\n')
    #             for line in lines:
    #                 if line.find('<url protocol=') != -1:
    #                     url = line.split(' >')[1].split('</url>')[0]
    #                     if url:
    #                         urls.append(url.strip())
    #                 if line.find('xmlns="') != -1:
    #                     url = line.split('xmlns="')[1].split('"')[0]
    #                     if url:
    #                         urls.append(url.strip())
    #                 if line.find('xmlns:mm0="') != -1:
    #                     url = line.split('xmlns:mm0="')[1].split('"')[0]
    #                     if url:
    #                         urls.append(url.strip())
    #
    #     _logger.debug('{0}: retrieving metalink list for section "{1}" succeeded'.format(self.get_name(), section))
    #     return urls



#
# Authors: Robert Abram <robert.abram@entpack.com>
#
# Copyright (C) 2015-2016 EntPack
# see file 'LICENSE' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from silentdune_client.models.json_base import JsonObject


class RemoteConfig(JsonObject):

    schema_version = None  # Should be 1 for now.
    hash_method = None  # Hashing method to use in verification, IE: SHA1, SHA256, ...
    rules = None  # Array of RemoteConfigRule objects

    def __init__(self, *args, **kwargs):
        super(RemoteConfig, self).__init__(args[0], kwargs)

        if self.schema_version == 1:
            self.rules = self.dict_to_obj_array(RemoteConfigRule_v1, self.rules)


class RemoteConfigRule_v1(JsonObject):

    host = None  # single IPv4 or IPv6 destination host
    mask = None  # netmask for host
    port = None  # destination network port
    protocol = None  # tcp, udp, icmp, ...
    uid = None  # for match "owner". IE: "-m owner --uid-owner 123"
    gid = None  # for match "owner". IE: "-m owner --gid-owner 123"



#
# Authors: Robert Abram <robert.abram@entpack.com>,
#
# Copyright (C) 2015-2016 EntPack
# see file 'LICENSE' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import hashlib
import logging
import platform
import socket
import sys
import time
from random import random

import pkg_resources
import requests
from silentdune_client.builders import iptables as ipt
from silentdune_client.models.global_preferences import GlobalPreferences
from silentdune_client.models.node import Node, NodeBundle
from silentdune_client.modules import BaseModule
from silentdune_client.modules import QueueTask
from silentdune_client.modules.comm.sd_server.connection import SilentDuneConnection, ConnStatus
from silentdune_client.modules.firewall.logging import SilentDuneClientLoggingModule, TASK_SUBSCRIBE_LOG_EVENTS, \
    TASK_LOG_EVENT
from silentdune_client.modules.firewall.manager import SilentDuneClientFirewallModule, \
    TASK_FIREWALL_INSERT_RULES, TASK_FIREWALL_DELETE_RULES
from silentdune_client.modules.firewall.manager.iptables_utils import create_iptables_tcp_ingress_egress_rule, \
    create_iptables_egress_ingress_rule
from silentdune_client.utils.misc import is_valid_ipv4_address, is_valid_ipv6_address

from silentdune_client.modules.firewall.manager.slots import Slots

_logger = logging.getLogger('sd-client')

TASK_SEND_SERVER_ALERT = 500

# Define the available Module classes.
module_list = {
    'Silent Dune Server': {
        'module': 'SilentDuneServerModule',
    },
}


class SilentDuneServerModule(BaseModule):
    """ Silent Dune Server Module """

    _enabled = False

    # Module properties
    _server = ''
    _port = 0
    _no_tls = False
    _bundle_name = ''
    _user = ''
    _password = ''

    _flat_rules = None

    # Server Connection
    _sds_conn = None
    _connected = False
    _connection_start = False

    # Server objects
    _node = None
    _bundle = None
    _node_bundle = None
    _bundle_machine_subsets = None
    _globals = None

    # Timed events.
    _t_connection_retry = 0  # Timed event property for retrying connection to server.
    _t_next_check = 0  # Timed event used to calculate next time we should contact the server.

    # Random number of seconds between -30 and 30 to make sure all nodes do not
    # contact the server at the same time.
    _t_random_seconds = int((random() - 0.5) * 60.0)
    _t_check_interval = 0

    priority = 30

    # Status
    _locked = False
    _startup = True

    def __init__(self):

        self._arg_name = 'server'  # Set argparser name
        self._config_section_name = 'server_module'  # Set configuration file section name

        # Enable multi-threading
        self.wants_processing_thread = True

        self._enabled = False

        try:
            self._version = pkg_resources.get_distribution(__name__).version
        except:
            self._version = 'unknown'

    def add_installer_arguments(self, parser):
        """
        Virtual Override
        Add our module's argparser arguments
        """

        # Create a argument group for our module
        group = parser.add_argument_group('server module', 'Silent Dune Server module')

        group.add_argument('--server-mod-enable', action='store_true', help=_('Enable the server module'))  # noqa

        group.add_argument(_('--server'), help=_('Silent Dune server network address (required)'),  # noqa
                        default=None, type=str, metavar='IP')  # noqa
        group.add_argument(_('--server-bundle'), help=_('Firewall bundle to use for this node (required)'),  # noqa
                        default=None, type=str, metavar='BUNDLE')  # noqa
        group.add_argument(_('--server-user'), help=_('Server admin user name (required)'),  # noqa
                        default=None, type=str, metavar='USER')  # noqa
        group.add_argument(_('--server-password'), help=_('Server admin password (required)'),  # noqa
                        default=None, type=str, metavar='PASSWORD')  # noqa
        group.add_argument(_('--server-no-tls'), help=_('Do not use a TLS connection'),  # noqa
                        default=False, action='store_true')  # noqa
        group.add_argument(_('--server-port'), help=_('Use alternate http port'),  # noqa
                        default=0, type=int, metavar='PORT')  # noqa

    def validate_arguments(self, args):
        """
        Virtual Override
        Validate command line arguments and save values to our configuration object.
        :param args: An argparse object.
        """
        # See if we have been enabled or not
        if '--server-mod-enable' not in sys.argv:
            return True

        if args.server_mod_enable:
            self._enabled = True

        if not args.server:
            print('sdc-install: argument --server is required.')
            return False
        if not args.server_bundle:
            print('sdc-install: argument --server-bundle is required.')
            return False
        if not args.server_user:
            print('sdc-install: argument --server-user is required.')
            return False
        if not args.server_password:
            print('sdc-install: argument --server-password is required.')
            return False

        # Check for valid IPv4 address
        if '.' in args.server:
            if not is_valid_ipv4_address(args.server):
                print('sdc-install: argument --server is invalid ip address')
                return False

        # Check for valid IPv6 address
        if ':' in args.server:
            if not is_valid_ipv6_address(args.server):
                print('sdc-install: argument --server is invalid ip address')
                return False

        self._server = args.server
        self._port = args.server_port
        self._no_tls = args.server_no_tls
        self._bundle_name = args.server_bundle

        # User and password are only used during the install process
        self._user = args.server_user
        self._password = args.server_password

        return True

    def validate_config(self, config):
        """
        Virtual Override
        Validate configuration file arguments and save values to our config object.
        :param config: A ConfigParser object.
        """
        # See if we are enabled or not
        try:
            self._enabled = True if config.get(self._config_section_name, 'enabled').lower() == 'yes' else False
        except:
            _logger.debug('{0} configuration section not found in configuration file.'.format(
                self._config_section_name))
            self._enabled = False

        # Only worry about the rest of the configuration items if we are enabled.
        if self._enabled:

            server = config.get(self._config_section_name, 'server')

            # Check for valid IPv4 address
            if '.' in server:
                if not is_valid_ipv4_address(server):
                    _logger.error('{0}: Config value for "server" is invalid ip address'.format(self.get_name()))
                    return False

            # Check for valid IPv6 address
            if ':' in server:
                if not is_valid_ipv6_address(server):
                    _logger.error('{0}: Config value for "server" is invalid ip address'.format(self.get_name()))
                    return False

            self._server = config.get(self._config_section_name, 'server')
            self._no_tls = False if config.get(self._config_section_name, 'use_tls').lower() == 'yes' else True

            # Make sure the port value is an integer and in range
            try:
                self._port = int(config.get(self._config_section_name, 'port'))
                if not (1 < self._port <= 65536):
                    _logger.error('{0}: port value out of range ({1})'.format(self.get_name(), self._port))
            except ValueError:
                self._port = 80 if self._no_tls else 443

            self._bundle_name = config.get(self._config_section_name, 'bundle')

        return True

    def prepare_config(self, config):
        """
        Virtual Override
        Return the configuration file structure. Any new configuration items should be added here.
        Note: The order should be reverse of the expected order in the configuration file.
        """

        config.set(self._config_section_name, 'enabled', 'yes' if self._enabled else 'no')
        config.set(self._config_section_name, 'server', self._server)
        config.set(self._config_section_name, 'port', self._port)
        config.set(self._config_section_name, 'use_tls', 'no' if self._no_tls else 'yes')
        config.set(self._config_section_name, 'bundle', self._bundle_name)

        config.set_comment(self._config_section_name, self._config_section_name,
                           _('; Silent Dune Server Module Configuration\n'))  # noqa
        config.set_comment(self._config_section_name, 'server',
                           _('; The Silent Dune management server to connect with.\n'))  # noqa
        config.set_comment(self._config_section_name, 'port',
                           _('; The port used by the management server. If no port is given this\n'  # noqa
                            '; node will use port 80 or 443 to connect to the management server\n'
                            '; depending on if the --no-tls option was used during the install.\n'))
        config.set_comment(self._config_section_name, 'use_tls',
                           _('; Use a secure connection when communicating with the management server.'))  # noqa
        config.set_comment(self._config_section_name, 'bundle',
                           _('; Name of the Bundle assigned to this node. Changing this value has\n'  # noqa
                             '; no affect. The client always uses the bundle information assigned\n'
                             '; by the server.'))

        return True

    def install_module(self):
        """
        Virtual Override
        Register and download our bundle information from the server.
        """

        if not self._enabled:
            return True

        self._sds_conn = SilentDuneConnection(self._server, self._no_tls, self._port)

        if not self._sds_conn.connect_with_password(self._user, self._password):
            return False

        self.get_global_preferences()

        if not self._register_node():
            return False

        if not self._get_rule_bundle():
            return False

        if not self._set_node_bundle():
            return False

        # TODO: Get and Upload adapter interface list to server
        # Note: It might be better to call ifconfig instead of using netifaces to get adapter info.

        # self._write_bundleset_to_file()

        return True

    def service_connect_to_server(self):
        """
        Attempt to connect to the silent dune server.
        :return:
        """

        self._connected = False
        self._connection_start = False

        self._sds_conn = SilentDuneConnection(self._server, self._no_tls, self._port)
        _logger.info('{0}: Server -> {1}:{2} use-tls: {3}'.format(
            self.get_name(), self._server, self._port, not self._no_tls))

        if self._sds_conn.connect_with_machine_id(self.node_info.machine_id):
            self._connected = True
            self._connection_start = True
            self._node, status_code = self._sds_conn.get_node_by_machine_id(self.node_info.machine_id)

            if self._node:
                self._t_check_interval = (self._node.polling_interval * 60) + self._t_random_seconds
            else:
                _logger.critical('{0}: !!! node information retrieval failed'.format(self.get_name()))
        else:
            # Check to see if
            _logger.warning('{0}: Failed to connect with server, will attempt reconnection.'.format(self.get_name()))

        return self._connected

    def service_startup(self):
        _logger.debug('{0}: module startup called'.format(self.get_name()))

        return True

    def service_shutdown(self):
        _logger.debug('{0}: module shutdown called'.format(self.get_name()))

        # Notify the server we are no longer active.
        if self._connected:
            self.update_server_node_info(active=False)

        return True

    def _add_server_access_rule(self):
        """
        Add a management server connection rule so we can connect
        :return:
        """

        rule = create_iptables_egress_ingress_rule(
            self._server, self._port, 'tcp', Slots.silentdune_server, transport=ipt.TRANSPORT_AUTO)

        # Notify the firewall module to reload the rules.
        task = QueueTask(TASK_FIREWALL_INSERT_RULES,
                         src_module=self.get_name(),
                         dest_module=SilentDuneClientFirewallModule().get_name(),
                         data=rule)
        self.send_parent_task(task)

    def process_loop(self):

        if self._startup:
            self._startup = False

            self._add_server_access_rule()
            time.sleep(3)  # Give the firewall time to add the rules before we attempt to connect

            self.service_connect_to_server()

            if self._sds_conn.status == ConnStatus.not_registered:
                _logger.error('{0} this node is not registered, unable to download rules'.format(self.get_name()))

        # If we are not registered, just return
        if self._sds_conn and self._sds_conn.status == ConnStatus.not_registered:
            return

        # If we are not connected to the server, try reconnecting every 60 seconds.
        if not self._connected:
            if self.timed_event('_t_connection_retry', 60):
                self._t_connection_retry = self.t_seconds
                self.service_connect_to_server()

        if self._connected:

            # Check to see if the connection to the server has just started.
            if self._connection_start:

                # Get the global preferences
                self.get_global_preferences()

                # Notify the server we are active now.
                self._node.active = True
                self.update_server_node_info(active=True)

                # Set our lock down mode value.
                self.set_lockdown_mode()

                # Update our firewall with rules from the server.
                self.update_node_firewall_rules_from_server()

                self._connection_start = False

                # Subscribe to log events
                task = QueueTask(TASK_SUBSCRIBE_LOG_EVENTS,
                                 src_module=self.get_name(),
                                 dest_module=SilentDuneClientLoggingModule().get_name())
                self.send_parent_task(task)

                _logger.debug('Next server check in {0} seconds.'.format(self._t_check_interval))

            # Check to see if the node rule bundle information has changed.
            if self.timed_event('_t_next_check', self._t_check_interval):
                _logger.debug('Next server check in {0} seconds.'.format(self._t_check_interval))

                self._node, status_code = self._sds_conn.get_node_by_machine_id(self.node_info.machine_id)

                # Check to see if we need to update our firewall rules bundle.
                if self._node.sync:
                    _logger.info('{0}: Found signal to update the firewall rules bundle.'.format(self.get_name()))

                    self.set_lockdown_mode()
                    self.update_node_firewall_rules_from_server()

                    # Update our information with the server.
                    self.update_server_node_info(sync=False)

    def process_task(self, task):

        if task:
            t_id = task.get_task_id()

            if t_id == TASK_LOG_EVENT:
                _logger.debug('{0}: received log event.'.format(self.get_name()))
                if self.node_info.platform == 'linux':

                    data = task.get_data()  # IPLogEvent object
                    self._sds_conn.iptables_send_alert(data)

    def set_lockdown_mode(self):
        """
        Set the lock down mode
        :return:
        """
        # Check locked down mode.
        if not self._locked and self._node.locked:
            _logger.warning('System is now in lock down mode.')
            self._locked = True

        if self._locked and not self._node.locked:
            _logger.warning('System is no longer in lock down mode.')
            self._locked = False

    def get_global_preferences(self):

        self._globals, status_code = self._sds_conn.get_global_preferences()

        if not globals or status_code != requests.codes.ok:
            _logger.error('Failed to download global preferences. Err: {0}'.format(status_code))
            self._globals = GlobalPreferences(lockdown_slot_level=120)
            return False

        return True

    def update_server_node_info(self, **kwargs):

        # _logger.debug(self._node.to_json())

        data = {u'machine_id': self._node.machine_id}  # machine_id is always required to update a node data record.

        if 'sync' in kwargs:
            data[u'sync'] = kwargs['sync']
            self._node.sync = kwargs['sync']

        if 'active' in kwargs:
            self._node.active = kwargs['active']
            data[u'active'] = kwargs['active']
        else:
            data[u'active'] = self._node.active

        reply, status_code = self._sds_conn.update_node(self._node.id, data)

        # _logger.debug(self._node.to_json())

        if not self._node or status_code != requests.codes.ok:
            _logger.error('{0}: Failed to update node information.'.format(self.get_name()))
            return False

        return True

    def update_node_firewall_rules_from_server(self):
        """
        Retrieve the node bundle rules from the server and send them to the Firewall module.
        :return:
        """

        # TODO: Save current bundle machine subsets and then look for any orphaned sets and remove them
        # from the firewall.

        if self._bundle_machine_subsets:
            # Until the orphan rule check is in place, just tell the firewall to delete all rules.
            task = QueueTask(TASK_FIREWALL_DELETE_RULES,
                             src_module=self.get_name(),
                             dest_module=SilentDuneClientFirewallModule().get_name(),
                             data=self._bundle_machine_subsets)
            self.send_parent_task(task)

        # Get updated bundle information.
        self._node_bundle, status_code = self._sds_conn.get_node_bundle_by_node_id(self._node.id)
        self._bundle, status_code = self._sds_conn.get_bundle_by_id(self._node_bundle.bundle)

        if not self._download_bundleset():
            _logger.error('{0}: Failed to download firewall rules bundle.'.format(self.get_name()))
            self._connected = False
            return False

        self._insert_server_connection_rule()

        if self._bundle_machine_subsets and len(self._bundle_machine_subsets) > 0:

            rules = self.flatten_rules(self._bundle_machine_subsets)

            # Check to see if we are in lockdown mode. If so filter out all
            if self._locked:
                data = list()
                for i in rules:
                    if i.slot <= self._globals.lockdown_slot_level or i.slot >= self._globals.rejection_slot_level:
                        data.append(i)
            else:
                data = rules

            # Notify the firewall module to reload the rules.
            task = QueueTask(TASK_FIREWALL_INSERT_RULES,
                             src_module=self.get_name(),
                             dest_module=SilentDuneClientFirewallModule().get_name(),
                             data=data)
            self.send_parent_task(task)

            # Notify the firewall module to reload the rules.
            # task = QueueTask(TASK_FIREWALL_RELOAD_RULES,
            #                  src_module=self.get_name(),
            #                  dest_module=SilentDuneClientFirewallModule().get_name())
            # self.send_parent_task(task)

            # Reset the node rule bundle check timer
            self._t_next_check = self.t_seconds

        else:
            _logger.error('{0}: No rules downloaded from server, unable to update firewall module.'.format(
                self.get_name()))
            return False

        return True

        # TODO: Port knocker event triggers.  https://github.com/moxie0/knockknock

        # Every 10 seconds, send the firewall module a QueueTask
        # if self._seconds_t > self._event_t and self._seconds_t % 4 == 0.0:
        #     self._event_t = self._seconds_t
        #
        #     _logger.debug('Sending {0} module a task.'.format(type(SilentDuneClientFirewallModule).__name__))
        #
        #     task = QueueTask(TASK_FIREWALL_RELOAD_RULES,
        #                      self.get_name(),
        #                      SilentDuneClientFirewallModule().get_name(),
        #                      None)
        #
        #     self.send_parent_task(task)

    def _register_node(self):
        """
        Contact the server to register this node with the server.
        """
        # Look for existing Node record.
        self._node, status_code = self._sds_conn.get_node_by_machine_id(self.node_info.machine_id)

        if status_code == requests.codes.ok and self._node and self._node.id:
            _logger.warning('{0}: Node already registered, using previously registered node info.'.format(
                self.get_name()))
            return True

        self.cwrite('{0}: Registering Node...  '.format(self.get_name()))

        node = Node(
            platform=self.node_info.firewall_platform,
            os=platform.system().lower(),
            dist=platform.dist()[0],
            dist_version=platform.dist()[1],
            hostname=socket.gethostname(),
            python_version=sys.version.replace('\n', ''),
            machine_id=self.node_info.machine_id,
            fernet_key='',  # Fernet.generate_key().decode('UTF-8'),
            polling_interval=self._globals.polling_interval
        )

        # Attempt to register this node on the SD server.
        self._node, status_code = self._sds_conn.register_node(node)

        if status_code != requests.codes.created or not self._node or self._node.id is None:
            self.cwriteline('[Failed]', 'Register Node failed, unknown reason.')
            return False

        self.cwriteline('[OK]', 'Node successfully registered.')

        return True

    def _get_rule_bundle(self):

        if self._bundle_name:

            self.cwrite('Looking up rule bundle...')

            self._bundle, status_code = self._sds_conn.get_bundle_by_name(self._bundle_name)

            if self._bundle and self._bundle.id > 0:
                self.cwriteline('[OK]', 'Found rule bundle.')
                return True

            self.cwriteline('[Failed]', 'Unable to find rule bundle named "{0}".'.format(self._bundle_name))

            # _logger.warning(_("Unable to find the rule bundle specified. The installer can try to lookup "  # noqa
            #                   "and use the default server rule bundle."))  # noqa
            #
            # self.cwrite(_('Do you want to use the server default rule bundle? [y/N]:'))  # noqa
            # result = sys.stdin.read(1)
            #
            # if result not in {'y', 'Y'}:
            #     _logger.debug('User aborting installation process.')
            #     return False

        self.cwrite('Looking up the server default rule bundle...')

        self._bundle, status_code = self._sds_conn.get_default_bundle()

        if not self._bundle or self._bundle.id is None:
            self.cwriteline('[Failed]', 'Default bundle lookup failed.')
            return False

        self.cwriteline('[OK]', 'Found default rule bundle.')

        return True

    def _set_node_bundle(self):

        self.cwrite('Setting Node rule bundle...')

        data = NodeBundle(node=self._node.id, bundle=self._bundle.id)

        self._node_bundle, status_code = self._sds_conn.create_node_bundle(data)

        if not self._node_bundle:
            self.cwriteline('[Failed]', 'Unable to set Node rule bundle.')
            return False

        self.cwriteline('[OK]', 'Node rule bundle successfully set.')
        return True

    def _download_bundleset(self):

        self.cwrite('Downloading bundle set rules...')

        # Get the machineset IDs assigned to the bundle
        self._bundle_machine_subsets, status_code = self._sds_conn.get_bundle_machine_subsets(self._node_bundle)
        if self._bundle_machine_subsets is None:
            self.cwriteline('[Failed]', 'No bundle machine subsets found.')
            return False

        return True

    def _insert_server_connection_rule(self):
        """
        Create machine subset to allow this node to access the Silent Dune server and then insert it into the bundle.
        :return:
        """
        ol = list()

        if self.node_info.firewall_platform == 'iptables':
            mss = create_iptables_tcp_ingress_egress_rule(
                self._server, self._port, slot=120, desc='Silent Dune Server Access')

            ol.append(mss)

        for ms in self._bundle_machine_subsets:
            ol.append(ms)

        self._bundle_machine_subsets = ol

    def flatten_rules(self, obj, level=0):
        """
        Takes nested rules and recursively un-winds them.
        :param obj: Rules object
        :param level: recursion level counter
        :return: rule list
        """

        if level == 0:
            self._flat_rules = list()

        try:

            # Test obj to see if it is really a rule
            if obj.to_json().encode('utf-8'):

                match = False

                for rule in self._flat_rules:
                    if hashlib.sha1(rule.to_json().encode('utf-8')).hexdigest() == \
                            hashlib.sha1(obj.to_json().encode('utf-8')).hexdigest():
                        match = True

                # Don't add rules that already exist in our flat list.
                if not match:
                    self._flat_rules.append(obj)

        except AttributeError:
            if obj and len(obj) > 0:
                for o in obj:
                    self.flatten_rules(o, (level+1))

        if level == 0:
            rules = self._flat_rules
            self._flat_rules = None  # clear list.
            return rules


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    Setup file for silentdune_client.

    This file was generated with PyScaffold 2.5, a tool that easily
    puts up a scaffold for your new Python project. Learn more under:
    http://pyscaffold.readthedocs.org/
"""

import os
import sys
from setuptools import setup, find_packages

__VERSION__ = "0.5.10"


def find_package_data_files(dirs):
    paths = []
    for directory in dirs:
        for (path, directories, filenames) in os.walk(directory):
            for filename in filenames:
                paths.append(os.path.join('..', path, filename))
    return paths


def setup_package():

    # Recursively gather all non-python module directories to be included in packaging.
    core_files = find_package_data_files([
        'silentdune_client/init',
        'silentdune_client/selinux',
        'silentdune_client/po',
    ])

    needs_sphinx = set(['build_sphinx', 'upload_docs']).intersection(set(sys.argv))
    sphinx = ['sphinx'] if needs_sphinx else []
    setup(
        name='SilentDune-Client',
        packages=find_packages(exclude=['tests']),
        package_data={
            'silentdune_client': core_files,
        },
        version=__VERSION__,
        description='Silent Dune FPA Client',
        author='Robert Abram',
        author_email='robert.abram@entpack.com',
        maintainer='Robert Abram',
        maintainer_email='robert.abram@entpack.com',
        url='https://github.com/EntPack/SilentDune-Client',
        download_url='https://github.com/EntPack/SilentDune-Client/tarball/0.1',
        keywords=['firewall', 'security'],  # arbitrary keywords
        classifiers=[
            'Development Status :: 4 - Beta',
            'Programming Language :: Python',
            'Environment :: Console',
            'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
            'Operating System :: POSIX :: Linux',
        ],

        install_requires=[
                             'setuptools',
                             # 'requests',
                             # 'cryptography',
                             # 'python-dateutil',
                             # 'dnspython',
                         ] + sphinx,

        tests_require=[
            'pytest',
            'pytest-runner',
            'pytest-pythonpath',
        ],

        entry_points={
            'console_scripts': [
                'sdc-install = silentdune_client.sdc_install:run',
                'sdc-firewall = silentdune_client.sdc_firewall:run',
            ],
        }
    )


if __name__ == "__main__":
    setup_package()



#
# Authors: Robert Abram <robert.abram@entpack.com>
#
# Copyright (C) 2015-2016 EntPack
# see file 'LICENSE' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import argparse
import gettext
import logging
import multiprocessing
import operator
import os
import signal
import sys
import time

from multiprocessing import Manager

from silentdune_client import modules
from silentdune_client.utils.log import setup_logging
from silentdune_client.utils.misc import node_info_dump

from silentdune_client.utils.configuration import ClientConfiguration
from silentdune_client.utils.daemon import Daemon

_logger = logging.getLogger('sd-client')


def run():

    class SDCDaemon(Daemon):

        # Node configuration information
        _args = None
        _config = None
        stopProcessing = False
        reload = False

        t_start = time.time()
        t_mod_check = 0

        def __init__(self, *args, **kwargs):

            self._args = kwargs.pop('args', None)

            super(SDCDaemon, self).__init__(*args, **kwargs)

        def startup_modules(self):

            # Get the path where this file is located.
            app_path = os.path.split(os.path.realpath(__file__))[0]
            # Get our package path and package name
            base_path, package_name = os.path.split(app_path)

            # Get loadable module list
            mods = modules.__load_modules__(base_path=base_path)
            active_mods = []  # List of modules marked as active.
            running_mods = []  # List of modules that are really running.

            # Set the configuration in each module.
            for mod in mods:
                mod.set_config(self._config)  # Set the configuration information in the module.

                # If the module is enabled, add it to the active_mods list.
                if mod.module_enabled():
                    active_mods.append(mod)
                else:
                    _logger.debug('Service: module {0} is disabled.'.format(mod.get_name()))

            pmanager = Manager()
            mqueue = pmanager.Queue()

            # Keep the created child processes.
            cprocs = dict()   # Dictionary of module process handlers.
            cqueues = dict()  # Dictionary of module Queue objects.
            mlist = list()    # List of module names.

            # Sort modules by the priority attribute so we can start them in the proper order.
            sorted_mods = sorted(active_mods, key=operator.attrgetter('priority'))

            for mod in sorted_mods:
                _logger.debug('Service: starting module {0}: ({1})'.format(mod.get_name(), mod.priority))

                if mod.service_startup() is False:
                    _logger.critical('Service: module ({0}) failed during startup.'.format(mod.get_name))
                    # sys.exit(1)
                    continue

                name = mod.get_name()

                running_mods.append(mod)
                mlist.append(name)

                cprocs[name] = None  # Add a place holder for the module process.

                # Setup thread for modules wanting a processing thread.
                if mod.wants_processing_thread:
                    # _logger.debug('Initializing thread for {0}.'.format(name))

                    cqueues[name] = multiprocessing.Queue()
                    cprocs[name] = multiprocessing.Process(
                        target=mod.process_handler, args=(cqueues[name], mqueue, mlist))
                    cprocs[name].start()

                    # Give the firewall manager time to setup the initial rules.
                    if name == 'SilentDuneClientFirewallModule':
                        time.sleep(2)

            return running_mods, pmanager, mqueue, cprocs, cqueues, mlist

        def check_module_state(self, mods, cprocs, force=False):
            """
            Check each module that has a thread and make sure it is still alive.
            :param mods:
            :return: False if all threads are running fine, True if failed module.
            """

            # We only want to do a check once a minute.
            time_t = int((time.time() - self.t_start))

            if (time_t > self.t_mod_check and time_t % 60.0 == 0.0) or force:
                self.t_mod_check = int((time.time() - self.t_start))

                # Check to see that module process threads are still running.
                _logger.debug('Service: checking module threads.')
                for mod in mods:
                    name = mod.get_name()
                    _logger.debug('{0}: checking module thread...'.format(name))
                    if name in cprocs and cprocs[name]:
                        if not cprocs[name].is_alive():
                            # TODO: Maybe restart the module?
                            _logger.critical('service: {0} module has unexpectedly stopped.'.format(name))
                            return True

            return False

        def terminate_modules(self, mods, cprocs, cqueues):
            """
            Shutdown modules.
            """
            for mod in mods:

                name = mod.get_name()

                if cprocs[name] and cprocs[name].is_alive():
                    _logger.debug('Service: signalling {0} module to stop processing.'.format(name))
                    cqueues[name].put(modules.QueueTask(modules.TASK_STOP_PROCESSING))
                    cqueues[name].close()
                    cqueues[name].join_thread()
                    cprocs[name].join()

        def run(self):

            _logger.debug('Service: setting signal handlers.')
            # Set SIGTERM signal Handler
            signal.signal(signal.SIGTERM, signal_term_handler)
            signal.signal(signal.SIGHUP, signal_hup_handler)

            _logger.info('Starting Silent Dune firewall.')

            # This loop allows for restarting and reloading the configuration after a SIGHUP signal has been received.
            while True:

                # Reset loop controllers
                self.stopProcessing = False
                self.reload = False

                # Read the local configuration file.
                self._config = ClientConfiguration(self._args.config).read_config()

                mods, pmanager, mqueue, cprocs, cqueues, mlist = self.startup_modules()

                # RUn main loop until we get an external signal.
                _logger.debug('Service: starting main processing loop.')
                while not self.stopProcessing:

                    if self.check_module_state(mods, cprocs):
                        # self.stopProcessing = True
                        break

                    # Check manage queue for any QueueTask object.
                    try:
                        task = mqueue.get_nowait()
                        _logger.debug('Service: forwarding task ({0}) from {1} to {2}'.format(
                            task.get_task_id(), task.get_sender(), task.get_dest_name()))

                        if task:
                            # Find the destination module and send task to it.
                            if not task.get_dest_name() or not cqueues[task.get_dest_name()]:
                                _logger.error('Service: task from {0} has unknown destination.'.format(
                                    task.get_sender()))

                            cqueues[task.get_dest_name()].put(task)
                    except:
                        pass

                    # Sleep.
                    time.sleep(0.25)

                # Stop all module processing threads
                _logger.debug('Service: terminating active modules...')

                self.terminate_modules(mods, cprocs, cqueues)

                # If we are not reloading, just shutdown.
                if not self.reload:
                    break

                _logger.debug('Service: reloading firewall.')

            _logger.info('Firewall shutdown complete.')

            # exit process
            sys.exit(0)

    def signal_term_handler(signal, frame):

        if not _daemon.stopProcessing:
            _logger.debug('Service: received SIGTERM signal.')
        _daemon.stopProcessing = True

    def signal_hup_handler(signal, frame):

        if not _daemon.reload:
            _logger.debug('Service: received SIGHUP signal.')
        _daemon.reload = True
        _daemon.stopProcessing = True

    setup_logging(_logger, '--debug' in sys.argv)

    os.environ['TMPDIR'] = '/var/run/silentdune'

    if not os.path.isdir(os.environ['TMPDIR']):
        os.makedirs(os.environ['TMPDIR'])

    # Setup i18n - Good for 2.x and 3.x python.
    kwargs = {}
    if sys.version_info[0] < 3:
        kwargs['unicode'] = True
    gettext.install('sdc_service', **kwargs)

    # Setup program arguments.
    parser = argparse.ArgumentParser(prog='sdc-firewall')
    parser.add_argument('-c', '--config', help=_('Use config file'), default=None, type=str)  # noqa
    parser.add_argument('--debug', help=_('Enable debug output'), default=False, action='store_true')  # noqa
    parser.add_argument('--nodaemon', help=_('Do not daemonize process'), default=False, action='store_true')  # noqa
    parser.add_argument('action', choices=('start', 'stop', 'restart'), default='')

    args = parser.parse_args()

    if os.getuid() != 0:
        print('sdc-firewall: error: must be run as root')
        sys.exit(4)

    # --nodaemon only valid with start action
    if args.nodaemon and args.action != 'start':
        print('sdc-firewall: error: --nodaemon option not valid with stop or restart action')
        sys.exit(1)

    # Read the local configuration file.
    config = ClientConfiguration(args.config).read_config()

    # Dump debug information
    if args.debug:
        node_info_dump(args)

    if not config:
        _logger.error('Invalid configuration file information, aborting.')
        sys.exit(1)

    # Do not fork the daemon process, run in foreground. For systemd service or debugging.
    if args.nodaemon:
        _daemon = SDCDaemon(args=args)
        _daemon.run()
    else:
        # Setup daemon object
        _daemon = SDCDaemon(
            os.path.split(config.get('settings', 'pidfile'))[0],
            '0o700',
            os.path.split(config.get('settings', 'pidfile'))[1],
            'root',
            'root',
            '/dev/null',
            '/dev/null',
            '/dev/null',
            args=args
        )

        if args.action == 'start':
            _logger.debug('Starting daemon.')
            _daemon.start()
        elif args.action == 'stop':
            _logger.debug('Stopping daemon.')
            _daemon.stop()
        elif args.action == 'restart':
            _logger.debug('Restarting daemon.')
            _daemon.restart()

    return 0


# --- Main Program Call ---
if __name__ == '__main__':
    sys.exit(run())



#
# Authors: Robert Abram <robert.abram@entpack.com>
#
# Copyright (C) 2015-2016 EntPack
# see file 'LICENSE' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import io
import logging
import operator
from datetime import datetime

from silentdune_client.models.json_base import JsonObject

_logger = logging.getLogger('sd-client')


class Bundle(JsonObject):
    """
    Represents the BundleSerializer json schema
    """
    id = None
    platform = None
    name = None
    descr = None
    notes = None
    default = False


class IPMachineSet(JsonObject):
    """
    Represents the IPMachineSetSerializer json schema
    """
    id = None
    chainset = None


class IPRulesFileWriter(object):
    """
    Object for writing iptables rules to a file
    """
    _chainsets = None

    def __init__(self, chainsets):
        self._chainsets = chainsets

    def _get_base_table_chains(self, table):
        """
        Return the builtin chains for the table specified and return the list.
        """
        chains = []

        if table == u'filter':
            chains = [u'INPUT', u'FORWARD', u'OUTPUT']
        elif table == u'nat':
            chains = [u'PREROUTING', u'OUTPUT', u'POSTROUTING']
        elif table == u'raw':
            chains = [u'PREROUTING', u'OUTPUT']
        elif table == u'mangle':
            chains = [u'PREROUTING', u'INPUT', u'FORWARD', u'OUTPUT', u'POSTROUTING']
        elif table == u'security':
            chains = [u'INPUT', u'FORWARD', u'OUTPUT']

        return chains

    def _output_rules(self, stream, version):
        """
        We will have one or more chainsets to loop through with each chainset having
        one or more IPChain objects.  We want to capture all of the IPChain objects,
        then loop through each one, match it to the iptables table we are looking at and then
        produce the data needed for writing the rules out.
        """

        # Loop through the iptables tables and write the rules for each one out.
        for table in [u'filter', u'nat', u'raw', u'mangle', u'security']:

            bases = self._get_base_table_chains(table)

            # See if there is at least rule for this table and the built-in chains for the table.
            if not self._check_for_table_rules(bases, table, version):
                continue

            # Start outputing the rulings in iptables save file format
            s = u'# Generated by SilentDune Client on {0}\n'.format(datetime.now().strftime(u'%a %b %d %H:%M:%S %Y'))
            s += u'*{0}\n'.format(table)  # Write out table name.
            stream.write(s)

            # Loop through and write out the builtin chain names for this table.
            for base in bases:
                s = u':{0} ACCEPT [0:0]\n'.format(base)
                stream.write(s)

            cnames = dict()

            # Loop through the chainsets and get the chain objects for each one.
            for chainset in iter(self._chainsets):
                if chainset.chains is not None:  # Make sure we don't have an empty IPChain list.
                    for chain in iter(chainset.chains):

                        # Only work with the IPChain objects that match the iptables table name.
                        if chain.name == table:

                            for base in bases:
                                cnames.update(chain.get_chain_names(chainset.slot, base, version))

            # Sort cnames by key (base iptables chain name), changes cnames from dict() to list of tuples.
            cnames = sorted(cnames.items(), key=operator.itemgetter(0))

            # Chain definitions need to be written in order of base chain name and slot id
            for base in bases:

                # Write out the chain definitions
                for key, val in cnames:
                    if val == base:
                        s = u':{0} - [0:0]\n'.format(key)
                        stream.write(s)

            # Chain jumps need to be written in order of base chain name and slot id
            for base in bases:

                # Write out the chain jumps
                for key, val in cnames:
                    if val == base:
                        s = u'-A {0} -j {1}\n'.format(val, key)  # val hold base chain name
                        stream.write(s)

            # Chain rules need to be written in order of base chain name and slot id
            for base in bases:

                # Write out the chain jumps, key contains the custom chain name.
                for key, val in cnames:
                    for chainset in iter(self._chainsets):
                        if chainset.chains is not None:  # Make sure we don't have an empty IPChain list.
                            chainset.write(stream, version, table, chainset.slot, key, base)

            s = u'COMMIT\n'
            stream.write(s)
            s = u'# Table {0} end\n'.format(table)
            stream.write(s)

            stream.flush()

    def _check_for_table_rules(self, chains, table, version):
        """
        We need to see if there is a IPRing object that matches the builtin chain and table
        in any of the chainsets we have. We are just looking for any match here.
        """

        # Loop through the iptables built-in table names.
        for c in chains:
            for cs in self._chainsets:
                if cs.check_for_machinesubset_rules(version, table, c):
                    return True  # Return true if we found a rule.

        return False

    def write_to_stream(self, stream, version):
        """
        Write iptables rules to a stream.
        :param stream: File stream stream
        :param version: Transport Version
        :return: True if successful or False if failure.
        """
        return self._output_rules(stream, version)

    def write_to_file(self, file, version):
        """
        Write iptables rules to a stream.
        :param file: String with file path and name to write iptables rules to.
        :param version: Transport Version
        :return: True if successful or False if failure.
        """
        with io.open(file, 'w') as stream:
            return self._output_rules(stream, version)


class IPMachineSubset(JsonObject):
    """
    Represents the IPMachineSubsetSerializer json schema
    """
    id = None  # PK value
    name = None
    descr = None
    notes = None
    platform = None
    slot = None
    sortId = None
    chains = None  # Array of IPChain objects

    def __init__(self, *args, **kwargs):
        super(IPMachineSubset, self).__init__(args[0], kwargs)
        self.chains = self.dict_to_obj_array(IPChain, self.chains)

    def check_for_machinesubset_rules(self, version, table, base):

        if self.chains is not None:
            for c in iter(self.chains):  # Call child objects write_chain method.
                if c.name == table:  # Only call if the table name matches.
                    if c.check_for_chain_rules(version, base):
                        return True  # If we found a rule, just return True immediately.
        return False

    def write(self, stream, version, table, slot, chain, base):

        if self.chains is not None:
            for o in iter(self.chains):  # Call child objects write method.
                if o.name == table:  # Only call if the table name matches.
                    o.write(stream, version, table, slot, chain, base)


class IPChain(JsonObject):
    """
    Represents the IPChainSerializer json schema
    """
    id = None  # PK value
    name = None  # iptables table name
    rings = None  # Array of IPRing objects

    def __init__(self, *args, **kwargs):
        super(IPChain, self).__init__(args[0], kwargs)
        self.rings = self.dict_to_obj_array(IPRing, self.rings)

    def get_chain_prefix(self, name):

        if name.lower() == u'prerouting':
            return u'PR'
        if name.lower() == u'postrouting':
            return u'PO'
        if name.lower() == u'forward':
            return u'FW'
        if name.lower() == u'input':
            return u'IN'
        if name.lower() == u'output':
            return u'OT'

        return None

    def get_chain_suffix(self, slot):

        if slot < 10:
            return u'_000{0}_{1}'.format(slot, self.id)
        elif slot < 100:
            return u'_00{0}_{1}'.format(slot, self.id)
        elif slot < 1000:
            return u'_0{0}_{1}'.format(slot, self.id)
        elif slot < 10000:
            return u'_{0}_{1}'.format(slot, self.id)
        else:
            _logger.error('{0}: Invalid iptables chain name by slot:{1}'.format(self.get_name(), slot))

        return None

    def get_chain_names(self, slot, cname, version):

        cnames = dict()

        # loop through the IPRing objects and build the iptables chain names.
        if self.rings is not None:
            for o in iter(self.rings):
                if o.version == version and o.name.lower() == cname.lower():
                    cnames[u'{0}{1}'.format(self.get_chain_prefix(o.name),
                                            self.get_chain_suffix(slot))] = cname  # save base chain name.

        return cnames

    def check_for_chain_rules(self, version, base):
        """
        Return True if there is a match by version and name
        """
        if self.rings is not None:
            for o in iter(self.rings):
                if o.version == version and o.name.upper() == base.upper():
                    return True

        return False

    def write(self, stream, version, table, slot, chain, base):

        if self.rings is not None:
            for o in iter(self.rings):  # Call child objects write method.
                n = u'{0}{1}'.format(self.get_chain_prefix(o.name), self.get_chain_suffix(slot))
                if o.version == version and o.name.lower() == base.lower() and chain == n:
                    o.write(stream, version, chain)


class IPRing(JsonObject):
    """
    Represents the IPRingSerializer json schema
    """
    id = None  # PK value
    name = None
    version = None  # Transport version ipv4 or ipv6
    rules = None  # Array of IPRule objects

    def __init__(self, *args, **kwargs):
        super(IPRing, self).__init__(args[0], kwargs)
        self.rules = self.dict_to_obj_array(IPRule, self.rules)

    def write(self, stream, version, chain):

        # Only write out rules if the versions match
        if version == self.version and self.rules is not None:
            for o in iter(self.rules):
                # iptables Chain name that the rule is attached to, matches the Chain name from write_chain.
                s = u'-A {0}'.format(chain)
                stream.write(s)

                # Write out the IPRule object's data
                o.write(stream)

                # Add line feed
                stream.write(u'\n')


class IPRule(JsonObject):
    """
    Represents the IPRuleSerializer json schema
    """
    id = None  # PK value
    enabled = None  # Is rule active
    sortId = None
    desc = None

    ifacein_name = None
    ifacein_invert = None
    ifaceout_name = None
    ifaceout_invert = None
    ip_protocol_name = None
    ip_protocol_invert = None
    source_address = None
    source_mask = None
    source_invert = None
    dest_address = None
    dest_mask = None
    dest_invert = None
    fragment = None
    fragment_invert = None

    matches = None  # Array of IPMatches objects
    jump = None  # Single IPJump object

    def __init__(self, *args, **kwargs):
        super(IPRule, self).__init__(args[0], kwargs)

        self.matches = self.dict_to_obj_array(IPMatch, self.matches)
        self.jump = None if self.jump is None else IPJump(self.jump)

    def write(self, stream):

        if self.fragment is True:
            s = u' {0}-f'.format(u'! ' if self.fragment_invert is True else u'')
            stream.write(s)

        if self.ifacein_name:
            s = u' -i{0} {1}'.format(u' !' if self.ifacein_invert is True else u'', self.ifacein_name)
            stream.write(s)

        if self.ifaceout_name:
            s = u' -o{0} {1}'.format(u' !' if self.ifaceout_invert is True else u'', self.ifaceout_name)
            stream.write(s)

        if self.ip_protocol_name:
            s = u' -p{0} {1}'.format(u' !' if self.ip_protocol_invert is True else u'', self.ip_protocol_name)
            stream.write(s)

        if self.source_address:
            if self.source_mask:
                s = u' -s{0} {1}/{2}'.format(u' !' if self.source_invert is True
                                             else u'', self.source_address, self.source_mask)
            else:
                s = u' -s{0} {1}'.format(u' !' if self.source_invert is True
                                         else u'', self.source_address)
            stream.write(s)

        if self.dest_address:
            if self.dest_mask:
                s = u' -d{0} {1}/{2}'.format(u' !' if self.dest_invert is True
                                             else u'', self.dest_address, self.dest_mask)
            else:
                s = u' -d{0} {1}'.format(u' !' if self.dest_invert is True
                                         else u'', self.dest_address)
            stream.write(s)

        if self.matches is not None:  # Call child objects write method.
            for o in iter(self.matches):
                o.write(stream)

        if self.jump is not None and self.jump.id is not None:
            self.jump.write(stream)


class IPMatch(JsonObject):
    """
    Represents the IPMatchSerializer json schema
    """
    id = None  # PK value
    name = None  # Match name
    options = None  # Array of IPMatchOptions objects

    def __init__(self, *args, **kwargs):
        super(IPMatch, self).__init__(args[0], kwargs)
        self.options = self.dict_to_obj_array(IPMatchOptions, self.options)

    def write(self, stream):
        s = u' -m {0}'.format(self.name)
        stream.write(s)

        for o in iter(self.options):  # Call child objects write method.
            o.write(stream)


class IPMatchOptions(JsonObject):
    """
    Represents the IPMatchOptionsSerializer json schema
    """
    id = None  # PK value
    option = None  # Option name
    value = None  # Option value
    invert = None  # Invert option value
    sortId = None

    def write(self, stream):
        s = u'{0} {1} {2}'.format(u' !' if self.invert is True else u'', self.option, self.value)
        stream.write(s)


class IPJump(JsonObject):
    """
    Represents the IPJumpSerializer json schema
    """
    id = None  # PK value
    target = None
    params = None  # Array of IPJumpOption objects

    def __init__(self, *args, **kwargs):
        super(IPJump, self).__init__(args[0], kwargs)
        self.params = self.dict_to_obj_array(IPJumpOptions, self.params)

    def write(self, stream):
        s = u' -j {0}'.format(self.target)
        stream.write(s)

        if self.params:
            for o in iter(self.params):  # Call child objects write method.
                o.write(stream)


class IPJumpOptions(JsonObject):
    """
    Represents the IPJumpOptionsSerializer json schema
    """
    id = None  # PK value
    name = None  # Option name
    value = None  # Option value

    def write(self, stream):

        # Identify any jump option parameters that need quotes around the value.
        str_parms = [
            '--log-prefix',
        ]

        if self.name in str_parms:
            s = u' {0} "{1}"'.format(self.name, self.value)
        else:
            s = u' {0} {1}'.format(self.name, self.value)

        stream.write(s)



#
# Authors: Robert Abram <robert.abram@entpack.com>
#
# Copyright (C) 2015-2016 EntPack
# see file 'LICENSE' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
import sys

_logger = logging.getLogger('sd-client')


class ConsoleBase(object):
    """
    Output messages to the console.
    """

    silent = False

    # http://stackoverflow.com/questions/11245381/formatting-console-output

    def cwrite(self, message, debug_msg=None):
        """
        Write a message to stdout or to debug logger with no linefeed.
        """

        if _logger.getEffectiveLevel() == logging.DEBUG:
            if debug_msg is None:
                _logger.debug(message)
            else:
                _logger.debug(debug_msg)
        elif not self.silent:
            # TODO: If silent, add the module name (prefix?) to the message
            sys.stdout.write(message)
            sys.stdout.flush()

    def cwriteline(self, message, debug_msg=None):
        """
        Write a message to stdout or to debug logger with linefeed.
        """

        if _logger.getEffectiveLevel() == logging.DEBUG:
            if debug_msg is None:
                _logger.debug(message)
            else:
                _logger.debug(debug_msg)
        elif not self.silent:
            # TODO: If silent, add the module name (prefix?) to the message
            print(message)
            sys.stdout.flush()



#
# Authors: Robert Abram <robert.abram@entpack.com>
#
# Copyright (C) 2015-2016 EntPack
# see file 'LICENSE' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
import os
import platform
import shutil
import socket
import subprocess
import sys

_logger = logging.getLogger('sd-client')


def which(program):
    """
    Find the path for a given program
    http://stackoverflow.com/questions/377017/test-if-executable-exists-in-python
    """

    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

    return None


def node_info_dump(args):
    """
    Output information about this machine.
    """
    # Basic system detections
    print('System = {0}'.format(platform.system()))

    # Current distribution
    print('Distribution = {0}'.format(platform.dist()[0]))
    print('Distribution Version = {0}'.format(platform.dist()[1]))

    # Python version
    print('Python Version: {0}'.format(sys.version.replace('\n', '')))

    print(args)


def rmdir(path):
    """
    Safely delete a path, make sure its not the filesystem root.
    :param path:
    :return:
    """
    if not path:
        return False

    path = os.path.realpath(path)

    if os.path.exists(path) and path != '/':
        shutil.rmtree(path)

    return True


def determine_config_root():
    """
    Determine where we are going to write the SD node configuration file.
    """

    home = os.path.expanduser('~')
    root_failed = False
    home_failed = False

    config_root = '/etc/silentdune/'

    # Check to see if the path already exist, if it does just return.
    if os.path.exists(config_root):
        return config_root

    # Test to see if we are running as root
    if os.getuid() == 0:
        test_file = os.path.join(config_root, 'test.tmp')

        try:
            os.makedirs(config_root)
            h = open(test_file, 'w')
            h.close()

            rmdir(config_root)

        except OSError:
            root_failed = True

    else:
        root_failed = True

    # If root access has failed, try the current user's home directory
    if root_failed:
        config_root = os.path.join(home, '.silentdune')

        # Check to see if the path already exist, if it does just return.
        if os.path.exists(config_root):
            return config_root

        test_file = os.path.join(config_root, 'test.tmp')

        try:
            os.makedirs(config_root)
            h = open(test_file, 'w')
            h.close()

            rmdir(config_root)

        except OSError:
            home_failed = True

    # Check if both locations failed.
    if root_failed and home_failed:
        _logger.critical('Unable to determine a writable configuration path for this node.')
        return None

    return config_root


def get_init_system():
    """
    Return the active init system on this node.
    :return:
    """
    # See if this system is an upstart setup.
    if which('initctl') is not None and os.path.isfile(which('initctl')):
        return 'upstart'

    # See if this system is a systemd setup.
    if which('systemctl') is not None and os.path.exists('/run/systemd/system'):
        return 'systemd'

    # See if this system is a sysvinit setup, must be after upstart and systemd detection.
    if which('service') is not None:
        return 'sysv'

    return None


def is_service_running(name):
    """
    Determine if the the specified service is running or active
    :param name:
    :return:
    """

    init = get_init_system()

    cmd = ['service', name, 'status']

    if init == 'systemd':
        cmd = ['systemctl', 'status', name]

    with open(os.devnull, 'w') as FNULL:
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=FNULL)
        stdoutdata, stderrdata = p.communicate()
        p.wait()

    if stderrdata is None:
        data = stdoutdata.decode('utf-8')
        if init == 'systemd' and 'Active: active' in data:
            return True
        elif 'running' in data:
            return True

    return False


def get_active_firewall():
    """
    Determine which firewall service is running on this system.
    :return: Firewall service name
    """
    if is_service_running('ufw'):
        return 'ufw'

    if is_service_running('firewalld'):
        return 'firewalld'

    if is_service_running('iptables'):
        return 'iptables'

    if is_service_running('sdc-firewall'):
        return 'sdc-firewall'

    return None


# http://stackoverflow.com/questions/319279/how-to-validate-ip-address-in-python
def is_valid_ipv4_address(address):
    """
    Validate an IPv4 IP address.
    :param address: IPv4 address with no netmask.
    :return: True if valid IPv4 IP address.
    """
    try:
        socket.inet_pton(socket.AF_INET, address)
    except AttributeError:  # no inet_pton here, sorry
        try:
            socket.inet_aton(address)
        except socket.error:
            return False
        return address.count('.') == 3
    except socket.error:  # not a valid address
        return False

    return True


def is_valid_ipv6_address(address):
    """
    Validate an IPv6 IP address.
    :param address: IPv6 address with no netmask.
    :return: True if valid IPv6 IP address.
    """
    try:
        socket.inet_pton(socket.AF_INET6, address)
    except socket.error:  # not a valid address
        return False
    return True


def validate_network_address(address):
    """
    Parse the network string for valid network address.
    :param address: Single or space delimited list of ipaddr/netmask values.
    :return: True if valid or False if incorrectly formatted.
    """

    if not address:
        return False

    for cidr in address.split(' '):

        try:
            ipaddr, bits = cidr.split('/')

            if '.' in bits and bits.count('.') == 3:  # See if we need to convert long notation to cidr notation.
                netmask = bits.split('.')
                binary_str = ''
                for octet in netmask:
                    if 0 <= int(octet) <= 255:
                        binary_str += bin(int(octet))[2:].zfill(8)
                    else:
                        return False
                bits = str(len(binary_str.strip('0')))

            bits = int(bits)

            if is_valid_ipv4_address(ipaddr) and 0 <= bits <= 32:
                continue

            if is_valid_ipv6_address(ipaddr) and 0 <= bits <= 128:
                continue

            return False

        except ValueError:
            return False

    return True



#
# Authors: Robert Abram <robert.abram@entpack.com>
#
# Copyright (C) 2015-2016 EntPack
# see file 'LICENSE' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from silentdune_client.models.json_base import JsonObject


class Node(JsonObject):
    """
    Represents the NodeSerializer json schema
    """
    id = None  # PK value
    platform = None  # Firewall platform, IE: iptables
    os = None  # System, IE: linux, windows, macos, freebsd, netbsd.
    dist = None  # Distribution Name.
    dist_version = None  # Distribution Version.
    hostname = None
    python_version = None
    machine_id = None  # Unique machine ID.
    last_connection = None  # Last connection datetime stamp.
    sync = None  # True if node needs to download new rule bundleset and update its information on the server.
    notes = None  # Notes about this node
    active = None  # Node firewall is actively running.
    locked = None  # Node should be in locked down or not.
    polling_interval = None  # A numeric value representing the number of minutes the node should check the server.
    fernet_key = None  # A fernet encryption key.


class NodeBundle(JsonObject):
    """
    Represents the NodeBundleSerializer json schema
    """
    id = None  # PK value
    node = None  # Node ID value
    bundle = None  # Bundle ID value


class NodeProcess(JsonObject):
    """
    Represents the NodeProcessSerializer json schema
    """
    id = None  # PK value
    node = None  # Node ID value
    name = None  # Process name from ps -ef
    category = None  # Process category name. IE: PostgreSQL Service, Apache Service.
    running = None  # True/False
    runningfor = None  # Process up time.
    ipaddr = None  # Bound IP addresses or 'all' for all available addresses.
    port = None  # Bound port number or range.
    lockable = None  # Firewall access rules can be removed to prevent traffic from reaching service.
    locked = None  # Firewall rules are removed for this process.
    delete = None  # This process database record is ready for deletion.

